from dataclasses import dataclass
from typing import Optional, Callable, Generic, TypeVar, Union

from .action import Action
from .condition import AndCondition, Condition
from .match_builder import merge_conditions
from .policy import RoutingPolicy, RoutingPolicyStatement
from .result import ResultType
from .statement_builder import StatementBuilder


class Route:
    def __init__(self, name: str):
        self.name = name
        self.statements: list[RoutingPolicyStatement] = []

    def __call__(
            self,
            *conditions: Condition,
            name: Optional[str] = None,
            number: Optional[int] = None,
    ) -> "StatementBuilder":
        statement = RoutingPolicyStatement(
            name=name,
            number=number,
            match=merge_conditions(AndCondition(*conditions)),
            then=Action(),
            result=ResultType.NEXT,
        )
        self.statements.append(statement)
        return StatementBuilder(statement=statement)


DeviceT = TypeVar("DeviceT")
RouteHandlerFunc = Callable[[DeviceT, Route], None]
Decorator = Callable[[RouteHandlerFunc[DeviceT]], RouteHandlerFunc[DeviceT]]


@dataclass
class Handler(Generic[DeviceT]):
    name: str
    func: RouteHandlerFunc[DeviceT]


class RouteMap(Generic[DeviceT]):
    def __init__(self) -> None:
        self.handlers: list[Handler[DeviceT]] = []
        self.submaps: list[RouteMap[DeviceT]] = []

    def __call__(
            self, func: Optional[RouteHandlerFunc[DeviceT]] = None, *, name: str = "",
    ) -> Union[RouteHandlerFunc[DeviceT], Decorator[DeviceT]]:
        def decorator(func: RouteHandlerFunc[DeviceT]) -> RouteHandlerFunc[DeviceT]:
            nonlocal name
            if not name:
                name = func.__name__
            self.handlers.append(Handler(name, func))
            return func

        if func is None:
            return decorator
        return decorator(func)

    def include(self, other: "RouteMap[DeviceT]") -> None:
        self.submaps.append(other)

    def apply(self, device: DeviceT) -> list[RoutingPolicy]:
        result: list[RoutingPolicy] = []

        for handler in self.handlers:
            route = Route(handler.name)
            handler.func(device, route)
            result.append(RoutingPolicy(route.name, route.statements))
        for submap in self.submaps:
            result.extend(submap.apply(device))
        return result



from collections.abc import Sequence
from dataclasses import dataclass

AS_PATH_FILTERS = {
    "ASP_EXAMPLE": [".*123456.*"],
}

IPV6_PREFIX_LISTS = {
    "IPV6_LIST_EXAMPLE": ["2a13:5941::/32"],
}


@dataclass(frozen=True)
class Community:
    values: Sequence[str]


@dataclass(frozen=True)
class ExtCommunity:
    values: Sequence[str]


COMMUNITIES = {
    "COMMUNITY_EXAMPLE_ADD": Community(["1234:1000"]),
    "COMMUNITY_EXAMPLE_REMOVE": Community(["12345:999"]),
}

EXT_COMMUNITIES = {
    "COMMUNITY_EXAMPLE_ADD": ExtCommunity(["1234:1000"]),
    "COMMUNITY_EXAMPLE_REMOVE": ExtCommunity(["12345:999"]),
}



from abc import ABC
from copy import copy
from enum import Enum
from typing import (
    TypeVar, Any, Annotated, get_origin, get_type_hints, get_args, Callable, Union, ClassVar, overload,
)


class MergeForbiddenError(Exception):
    pass


class Special(Enum):
    NOT_SET = "<NOT SET>"


T = TypeVar("T")


class Merger(ABC):
    def __call__(self, name: str, x: Union[T, Special], y: Union[T, Special]) -> Union[T, Special]:
        if x is Special.NOT_SET:
            return y
        if y is Special.NOT_SET:
            return x
        return self._merge(name, x, y)

    def _merge(self, name: str, x: T, y: T) -> T:
        raise NotImplementedError


class UseFirst(Merger):
    def _merge(self, name: str, x: T, y: T) -> T:
        return x


class UseLast(Merger):
    def _merge(self, name: str, x: T, y: T) -> T:
        return y


class Forbid(Merger):
    def _merge(self, name: str, x: T, y: T) -> T:
        raise MergeForbiddenError(f"Override is forbidden for field {name}")


class ForbidChange(Merger):
    def _merge(self, name: str, x: T, y: T) -> T:
        if x == y:
            return x
        raise MergeForbiddenError(
            f"Override with different value is forbidden for field {name}:\n"
            f"Old: {x}\n"
            f"New: {y}"
        )


class Concat(Merger):
    def _merge(self, name: str, x: T, y: T) -> T:
        return x + y  # type: ignore[operator]


class Merge(Merger):
    def _merge(self, name: str, x: "ModelT", y: "ModelT") -> "ModelT":  # type: ignore[override]
        return merge(x, y)


class DictMerge(Merger):
    def __init__(self, value_merger: Merger = Forbid()):
        self.value_merger = value_merger

    def _merge(self, name: str, x: dict, y: dict) -> dict:  # type: ignore[override]
        result = copy(x)
        for key, value in y.items():
            if key in result:
                result[key] = self.value_merger(key, result[key], value)
            else:
                result[key] = value
        return result


class ApplyFunc(Merger):
    def __init__(self, func: Callable):
        self.func = func

    def __call__(self, name: str, x: Union[T, Special], y: Union[T, Special]) -> Union[T, Special]:
        if x is Special.NOT_SET:
            return y
        if y is Special.NOT_SET:
            return x
        return self.func(x, y)


def _get_merger(hint: Any):
    if get_origin(hint) is not Annotated:
        return ForbidChange()
    for arg in get_args(hint):
        if isinstance(arg, Merger):
            return arg
    return ForbidChange()


class BaseMeshModel:
    _field_mergers: ClassVar[dict[str, Merger]]

    def __init__(self, **kwargs):
        if extra_keys := (kwargs.keys() - self._field_mergers.keys()):
            raise ValueError(f"Extra arguments: {extra_keys}")
        self.__dict__.update(kwargs)

    def unset_attrs(self):
        return type(self)._field_mergers.keys() - vars(self).keys()

    def __repr__(self):
        return f"{self.__class__.__name__}(" + ", ".join(
            f"{key}={value}" for key, value in vars(self).items()
        ) + ")"

    def __init_subclass__(cls, **kwargs):
        cls._field_mergers = {
            field: _get_merger(hint)
            for field, hint in get_type_hints(cls, include_extras=True).items()
            if get_origin(hint) is not ClassVar
        }

    def __setattr__(self, key, value):
        if key not in self._field_mergers:
            raise AttributeError(f"{self.__class__.__name__} has no field {key}")
        super().__setattr__(key, value)


ModelT = TypeVar("ModelT", bound=BaseMeshModel)


def _merge(a: ModelT, b: BaseMeshModel) -> ModelT:
    result = copy(a)
    for attr_name, merger in a._field_mergers.items():
        new_value = merger(
            attr_name,
            getattr(a, attr_name, Special.NOT_SET),
            getattr(b, attr_name, Special.NOT_SET),
        )
        if new_value is not Special.NOT_SET:
            setattr(result, attr_name, new_value)
    return result


@overload
def merge(first: Special, second: ModelT, /, *others: BaseMeshModel) -> ModelT:
    ...


@overload
def merge(first: ModelT, /, *others: BaseMeshModel) -> ModelT:
    ...


@overload
def merge(first: Special, /) -> Special:
    ...


def merge(first: Any, /, *others: Any) -> Any:
    if first is Special.NOT_SET:
        if not others:
            return Special.NOT_SET
        return merge(*others)
    for second in others:
        first = _merge(first, second)
    return first


class KeyDefaultDict(dict):
    def __init__(self, factory: Callable[[str], Any]):
        super().__init__()
        self.factory = factory

    def __missing__(self, key):
        x = self[key] = self.factory(key)
        return x



from typing import Optional

from annet.annlib.netdev.devdb import parse_hw_model

from .dump import DumpableView


class HardwareLeaf(DumpableView):
    def __new__(cls, *_, **__):
        obj = super(HardwareLeaf, cls).__new__(cls)
        obj.__path = ()
        obj.__true_sequences = set()
        obj.__false_sequences = set()
        return obj

    def __init__(self, path, true_sequences, false_sequences):
        self.__path = path
        self.__true_sequences = true_sequences
        self.__false_sequences = false_sequences

    def __bool__(self):
        if len(self.__path) == 0 or self.__path in self.__true_sequences:
            return True
        elif self.__path in self.__false_sequences:
            return False
        raise AttributeError("HW: " + ".".join(self.__path))

    def __getattr__(self, name):
        path = self.__path + (name,)
        if path in self.__true_sequences or path in self.__false_sequences:
            return HardwareLeaf(path, self.__true_sequences, self.__false_sequences)
        try:
            return self.__dict__[name]
        except KeyError:
            raise AttributeError("HW: " + ".".join(path))

    def __str__(self):
        return str(" | ".join(".".join(x) for x in self.__true_sequences))

    __repr__ = __str__

    def dump(self, prefix, **kwargs):  # pylint: disable=arguments-differ
        ret = super().dump(prefix, **kwargs)
        seen = set()
        for seq in sorted(self.__true_sequences, key=len, reverse=True):
            if any(name in seen for name in seq):
                continue
            seen.update(seq)
            ret.append("%s.%s = True" % (prefix, ".".join(seq)))
        return ret


class HardwareView(HardwareLeaf):
    def __init__(self, hw_model, sw_version):
        (true_sequences, false_sequences) = parse_hw_model(hw_model or "")
        super().__init__((), true_sequences, false_sequences)
        self.model = hw_model or ""
        self._soft = sw_version or ""

    @property
    def vendor(self) -> Optional[str]:
        return hw_to_vendor(self)

    @property
    def soft(self) -> str:
        return self._soft

    def __hash__(self):
        return hash(self.model)

    def __eq__(self, other):
        return self.model == other.model


def hw_to_vendor(hw: HardwareView) -> Optional[str]:
    if hw.Nexus:
        return "nexus"
    elif hw.Cisco:
        return "cisco"
    elif hw.OptiXtrans:
        return "optixtrans"
    elif hw.Huawei:
        return "huawei"
    elif hw.Juniper:
        return "juniper"
    elif hw.Arista:
        return "arista"
    elif hw.PC:
        return "pc"
    elif hw.Nokia:
        return "nokia"
    elif hw.RouterOS:
        return "routeros"
    elif hw.Aruba:
        return "aruba"
    elif hw.Ribbon:
        return "ribbon"
    elif hw.H3C:
        return "h3c"
    elif hw.B4com:
        return "b4com"
    return None


def vendor_to_hw(vendor):
    hw = HardwareView(
        {
            "cisco": "Cisco",
            "catalyst": "Cisco Catalyst",
            "nexus": "Cisco Nexus",
            "huawei": "Huawei",
            "optixtrans": "Huawei OptiXtrans",
            "juniper": "Juniper",
            "arista": "Arista",
            "pc": "PC",
            "nokia": "Nokia",
            "aruba": "Aruba",
            "routeros": "RouterOS",
            "ribbon": "Ribbon",
            "h3c": "H3C",
            "b4com": "B4com",
        }.get(vendor.lower(), vendor),
        None,
    )
    return hw


def lag_name(hw: HardwareView, nlagg: int) -> str:
    if hw.Huawei:
        return f"Eth-Trunk{nlagg}"
    if hw.Cisco:
        return f"port-channel{nlagg}"
    if hw.Nexus:
        return f"port-channel{nlagg}"
    if hw.Arista:
        return f"Port-Channel{nlagg}"
    if hw.Juniper:
        return f"ae{nlagg}"
    if hw.Nokia:
        return f"lag-{nlagg}"
    if hw.PC.Whitebox:
        return f"bond{nlagg}"
    if hw.PC:
        return f"lagg{nlagg}"
    if hw.Nokia:
        return f"lagg-{nlagg}"
    raise NotImplementedError(hw)


def svi_name(hw: HardwareView, num: int) -> str:
    if hw.Juniper:
        return f"irb.{num}"
    elif hw.Huawei:
        return f"Vlanif{num}"
    else:
        return f"vlan{num}"



import re
from collections.abc import Callable
from dataclasses import dataclass
from types import SimpleNamespace
from typing import Any, Sequence, Optional

MatchedArgs = SimpleNamespace


def identity(x: Any) -> Any:
    return x


class MatchExpr:
    def __init__(self, expr: Callable[[Any], Any] = identity):
        self.expr = expr

    def __getattr__(self, item: str):
        return MatchExpr(lambda x: getattr(self.expr(x), item))

    def __getitem__(self, item: Any):
        return MatchExpr(lambda x: self.expr(x)[item])

    def __eq__(self, other) -> "MatchExpr":  # type: ignore[override]  # https://github.com/python/mypy/issues/5951
        if isinstance(other, MatchExpr):
            return MatchExpr(lambda x: self.expr(x) == other.expr(x))
        else:
            return MatchExpr(lambda x: self.expr(x) == other)

    def __ne__(self, other) -> "MatchExpr":  # type: ignore[override]  # https://github.com/python/mypy/issues/5951
        if isinstance(other, MatchExpr):
            return MatchExpr(lambda x: self.expr(x) != other.expr(x))
        else:
            return MatchExpr(lambda x: self.expr(x) != other)

    def __lt__(self, other) -> "MatchExpr":
        if isinstance(other, MatchExpr):
            return MatchExpr(lambda x: self.expr(x) < other.expr(x))
        else:
            return MatchExpr(lambda x: self.expr(x) < other)

    def __gt__(self, other) -> "MatchExpr":
        if isinstance(other, MatchExpr):
            return MatchExpr(lambda x: self.expr(x) > other.expr(x))
        else:
            return MatchExpr(lambda x: self.expr(x) > other)

    def __le__(self, other) -> "MatchExpr":
        if isinstance(other, MatchExpr):
            return MatchExpr(lambda x: self.expr(x) <= other.expr(x))
        else:
            return MatchExpr(lambda x: self.expr(x) <= other)

    def __ge__(self, other) -> "MatchExpr":
        if isinstance(other, MatchExpr):
            return MatchExpr(lambda x: self.expr(x) >= other.expr(x))
        else:
            return MatchExpr(lambda x: self.expr(x) >= other)

    def __or__(self, other: "MatchExpr") -> "MatchExpr":
        return MatchExpr(lambda x: self.expr(x) or other.expr(x))

    def __and__(self, other: "MatchExpr") -> "MatchExpr":
        return MatchExpr(lambda x: self.expr(x) and other.expr(x))

    def cast_(self, type_: Callable[[Any], Any]) -> "MatchExpr":
        return MatchExpr(lambda x: type_(self.expr(x)))

    def in_(self, value: Any) -> "MatchExpr":
        if isinstance(value, MatchExpr):
            return MatchExpr(lambda x: self.expr(x) in value.expr(x))
        else:
            return MatchExpr(lambda x: self.expr(x) in value)


Match = MatchExpr()[0]
Left = MatchExpr()[0]
Right = MatchExpr()[1]


class PeerNameTemplate:
    def __init__(self, raw_str):
        self._str = str(raw_str)
        self._regex, self._types = self._compile(self._str)

    def __str__(self):
        return self._str

    @staticmethod
    def _compile(value: str) -> tuple[re.Pattern[str], dict[str, type]]:
        int_groups = re.findall(r"{(?P<group_name>\w+)}", value)
        # '{name}'  -> (?P<name>\d+)
        regex_string = re.sub(r"{(?P<group_name>\w+)}", r"(?P<\g<group_name>>\\d+)", value)
        # '{name:regex}' -> (?P<name>regex)
        regex_string = re.sub(
            r"{(?P<group_name>\w+):(?P<custom_regex>.*?)}", r"(?P<\g<group_name>>\g<custom_regex>)",
            regex_string
        )
        pattern = re.compile(regex_string)
        types: dict[str, type] = {
            name: (int if name in int_groups else str)
            for name in pattern.groupindex
        }
        return pattern, types

    def match(self, hostname: str) -> Optional[dict[str, str]]:
        reg_match = self._regex.fullmatch(hostname)
        if reg_match:
            return {
                key: self._types[key](value)
                for key, value in reg_match.groupdict().items()
            }
        return None


def match_safe(match_expressions: Sequence[MatchExpr], value: Any) -> bool:
    for matcher in match_expressions:
        try:
            res = matcher.expr(value)
            if not bool(res):
                return False
        except (TypeError, ValueError, AttributeError, KeyError, IndexError):
            return False
    return True


@dataclass
class SingleMatcher:
    def __init__(self, rule: str, match_expressions: Sequence[MatchExpr]):
        self.rule = PeerNameTemplate(rule)
        self.match_expressions = match_expressions

    def match_one(self, host: str) -> Optional[MatchedArgs]:
        data = self.rule.match(host)
        if data is None:
            return None
        args = MatchedArgs(**data)
        if not match_safe(self.match_expressions, (args,)):
            return None
        return args


@dataclass
class PairMatcher:
    def __init__(self, left_rule: str, right_rule: str, match_expressions: Sequence[MatchExpr]):
        self.left_rule = PeerNameTemplate(left_rule)
        self.right_rule = PeerNameTemplate(right_rule)
        self.match_expressions = match_expressions

    def match_pair(self, left: str, right: str) -> Optional[tuple[MatchedArgs, MatchedArgs]]:
        left_args = self._match_host(self.left_rule, left)
        if left_args is None:
            return None
        right_args = self._match_host(self.right_rule, right)
        if right_args is None:
            return None
        if not match_safe(self.match_expressions, (left_args, right_args)):
            return None
        return left_args, right_args

    def _match_host(self, rule: PeerNameTemplate, host: str) -> Optional[MatchedArgs]:
        data = rule.match(host)
        if data is None:
            return None
        return MatchedArgs(**data)



import logging
import logging.config
import os
import pkgutil
import sys
from argparse import SUPPRESS, Namespace

import colorama
import yaml
from annet.annlib.errors import (  # pylint: disable=wrong-import-position
    DeployCancelled,
    ExecError,
)
from contextlog import patch_logging, patch_threading
from valkit.python import valid_logging_level

import annet.argparse


__all__ = ("DeployCancelled", "ExecError")

DEBUG2_LEVELV_NUM = 9


def fill_base_args(parser: annet.argparse.ArgParser, pkg_name: str, logging_config: str):
    parser.add_argument("--log-level", default="WARN", type=valid_logging_level,
                        help="Уровень детализации логов (DEBUG, DEBUG2 (with comocutor debug), INFO, WARN, CRITICAL)")
    parser.add_argument("--pkg_name", default=pkg_name, help=SUPPRESS)
    parser.add_argument("--logging_config", default=logging_config, help=SUPPRESS)


def init_logging(options: Namespace):
    patch_logging()
    patch_threading()
    logging.captureWarnings(True)
    logging_config = yaml.safe_load(pkgutil.get_data(options.pkg_name, options.logging_config))
    if options.log_level is not None:
        logging_config.setdefault("root", {})
        logging_config["root"]["level"] = options.log_level
    logging.addLevelName(DEBUG2_LEVELV_NUM, "DEBUG2")
    logging.config.dictConfig(logging_config)


def init(options: Namespace):
    init_logging(options)

    # Отключить colorama.init, если стоит env-переменная. Нужно в тестах
    if os.environ.get("ANN_FORCE_COLOR", None) not in [None, "", "0", "no"]:
        colorama.init = lambda *_, **__: None
    colorama.init()

    # Workaround for Python 3.8.0: https://bugs.python.org/issue38529
    import asyncio.streams
    if hasattr(asyncio.streams.StreamReaderProtocol, "_on_reader_gc"):
        asyncio.streams.StreamReaderProtocol._on_reader_gc = lambda *args, **kwargs: None  # pylint: disable=protected-access


def assert_python_version():
    if sys.version_info < (3, 10, 0):
        sys.stderr.write("Error: you need python 3.10.0 or higher\n")
        sys.exit(1)



from dataclasses import dataclass
from logging import getLogger
from typing import Annotated, Callable, Optional, Union

from annet.bgp_models import Peer, GlobalOptions
from annet.storage import Device, Storage
from .basemodel import merge, BaseMeshModel, Merge, UseLast, MergeForbiddenError
from .device_models import GlobalOptionsDTO
from .models_converter import to_bgp_global_options, to_bgp_peer, InterfaceChanges, to_interface_changes
from .peer_models import DirectPeerDTO, IndirectPeerDTO, VirtualLocalDTO, VirtualPeerDTO
from .registry import (
    DirectPeer,
    GlobalOptions as MeshGlobalOptions,
    IndirectPeer,
    MeshRulesRegistry,
    MeshSession,
    VirtualLocal,
    VirtualPeer,
)

logger = getLogger(__name__)


@dataclass
class MeshExecutionResult:
    global_options: GlobalOptions
    peers: list[Peer]


@dataclass(frozen=True)
class PeerKey:
    fqdn: str
    addr: str
    vrf: str


class Pair(BaseMeshModel):
    local: Annotated[Union[DirectPeerDTO, IndirectPeerDTO], Merge()]
    connected: Annotated[Union[DirectPeerDTO, IndirectPeerDTO], Merge()]
    device: Annotated[Device, UseLast()]


class VirtualPair(BaseMeshModel):
    local: Annotated[VirtualLocalDTO, Merge()]
    connected: Annotated[VirtualPeerDTO, Merge()]


class MeshExecutor:
    def __init__(
            self,
            registry: MeshRulesRegistry,
            storage: Storage,
    ):
        self._registry = registry
        self._storage = storage

    def _execute_globals(self, device: Device) -> GlobalOptionsDTO:
        global_opts = GlobalOptionsDTO()
        for rule in self._registry.lookup_global(device.fqdn):
            handler_name = self._handler_name(rule.handler)
            rule_global_opts = MeshGlobalOptions(rule.match, device)
            logger.debug("Running device handler: %s", handler_name)
            rule.handler(rule_global_opts)
            try:
                global_opts = merge(global_opts, rule_global_opts)
            except MergeForbiddenError as e:
                raise ValueError(
                    f"Handler `{handler_name}` global options conflicting with "
                    f"previously loaded for device `{device.fqdn}`:\n" + str(e)
                ) from e
        return global_opts

    def _handler_name(self, handler: Callable) -> str:
        try:
            return f"{handler.__module__}.{handler.__qualname__}"
        except AttributeError:
            return str(handler)

    def _execute_direct(self, device: Device) -> list[Pair]:
        # we can have multiple rules for the same pair
        # we merge them according to remote fqdn
        neighbor_peers: dict[PeerKey, Pair] = {}
        # TODO batch resolve
        for rule in self._registry.lookup_direct(device.fqdn, device.neighbours_fqdns):
            session = MeshSession()
            handler_name = self._handler_name(rule.handler)
            logger.debug("Running direct handler: %s", handler_name)
            if rule.direct_order:
                neighbor_device = self._storage.make_devices([rule.name_right])[0]
                peer_device = DirectPeer(rule.match_left, device, [])
                peer_neighbor = DirectPeer(rule.match_right, neighbor_device, [])
            else:
                neighbor_device = self._storage.make_devices([rule.name_left])[0]
                peer_neighbor = DirectPeer(rule.match_left, neighbor_device, [])
                peer_device = DirectPeer(rule.match_right, device, [])

            interfaces = self._storage.search_connections(device, neighbor_device)
            for local_port, remote_port in interfaces:
                peer_device.ports.append(local_port.name)
                peer_neighbor.ports.append(remote_port.name)

            if rule.direct_order:
                rule.handler(peer_device, peer_neighbor, session)
            else:
                rule.handler(peer_neighbor, peer_device, session)

            try:
                neighbor_dto = merge(DirectPeerDTO(), peer_neighbor, session)
            except MergeForbiddenError as e:
                raise ValueError(
                    f"Handler `{handler_name}` provided session data conflicting with "
                    f"peer data for device `{neighbor_device.fqdn}`:\n" + str(e)
                ) from e
            try:
                device_dto = merge(DirectPeerDTO(), peer_device, session)
            except MergeForbiddenError as e:
                raise ValueError(
                    f"Handler `{handler_name}` provided session data conflicting with "
                    f"peer data for device `{device.fqdn}`:\n" + str(e)
                ) from e

            pair = Pair(local=device_dto, connected=neighbor_dto, device=neighbor_device)
            addr = getattr(neighbor_dto, "addr", None)
            if addr is None:
                raise ValueError(f"Handler `{handler_name}` returned no peer addr")
            peer_key = PeerKey(
                fqdn=neighbor_device.fqdn,
                addr=addr,
                vrf=getattr(neighbor_dto, "vrf", "")
            )
            try:
                if peer_key in neighbor_peers:
                    pair = merge(neighbor_peers[peer_key], pair)
            except MergeForbiddenError as e:
                if rule.direct_order:
                    pair_names = device.fqdn, neighbor_device.fqdn
                else:
                    pair_names = neighbor_device.fqdn, device.fqdn
                raise ValueError(
                    f"Handler `{handler_name}` provides data conflicting with "
                    f"previously loaded for device pair {pair_names} "
                    f"with addr={peer_key.addr}, vrf{peer_key.vrf}:\n" + str(e)
                ) from e
            neighbor_peers[peer_key] = pair
        return list(neighbor_peers.values())

    def _execute_virtual(self, device: Device) -> list[VirtualPair]:
        virtual_peers: list[VirtualPair] = []
        for rule in self._registry.lookup_virtual(device.fqdn):
            for order_number in rule.num:
                handler_name = self._handler_name(rule.handler)
                logger.debug("Running direct handler: %s", handler_name)
                session = MeshSession()
                peer_device = VirtualLocal(rule.match, device)
                peer_virtual = VirtualPeer(num=order_number)

                rule.handler(peer_device, peer_virtual, session)

                try:
                    virtual_dto = merge(VirtualPeerDTO(), peer_virtual, session)
                except MergeForbiddenError as e:
                    raise ValueError(
                        f"Handler `{handler_name}` provided session data conflicting with "
                        f"virtual peer data for device `{device.fqdn}` and num={order_number}:\n" + str(e)
                    ) from e
                try:
                    device_dto = merge(VirtualLocalDTO(), peer_device, session)
                except MergeForbiddenError as e:
                    raise ValueError(
                        f"Handler `{handler_name}` provided session data conflicting with "
                        f"peer data for device `{device.fqdn}`:\n" + str(e)
                    ) from e

                if not hasattr(device_dto, "svi"):
                    raise ValueError(
                        f"Handler `{handler_name}` did not provide `svi` number. "
                        "Virtual peer must be connected to SVI interface."
                    )
                pair = VirtualPair(local=device_dto, connected=virtual_dto)
                virtual_peers.append(pair)
        return virtual_peers

    def _execute_indirect(self, device: Device, all_fqdns: list[str]) -> list[Pair]:
        # we can have multiple rules for the same pair
        # we merge them according to remote fqdn
        connected_peers: dict[PeerKey, Pair] = {}
        for rule in self._registry.lookup_indirect(device.fqdn, all_fqdns):
            session = MeshSession()
            handler_name = self._handler_name(rule.handler)
            logger.debug("Running indirect handler: %s", handler_name)
            if rule.direct_order:
                connected_device = self._storage.make_devices([rule.name_right])[0]
                peer_device = IndirectPeer(rule.match_left, device)
                peer_connected = IndirectPeer(rule.match_right, connected_device)
                rule.handler(peer_device, peer_connected, session)
            else:
                connected_device = self._storage.make_devices([rule.name_left])[0]
                peer_connected = IndirectPeer(rule.match_left, connected_device)
                peer_device = IndirectPeer(rule.match_right, device)
                rule.handler(peer_connected, peer_device, session)

            try:
                connected_dto = merge(IndirectPeerDTO(), peer_connected, session)
            except MergeForbiddenError as e:
                raise ValueError(
                    f"Handler `{handler_name}` provided session data conflicting with "
                    f"peer data for device `{connected_device.fqdn}`:\n" + str(e)
                ) from e
            try:
                device_dto = merge(IndirectPeerDTO(), peer_device, session)
            except MergeForbiddenError as e:
                raise ValueError(
                    f"Handler `{handler_name}` provided session data conflicting with "
                    f"peer data for device `{device.fqdn}`:\n" + str(e)
                ) from e

            pair = Pair(local=device_dto, connected=connected_dto, device=connected_device)
            addr = getattr(connected_dto, "addr", None)
            if addr is None:
                raise ValueError(f"Handler `{handler_name}` returned no peer addr")
            peer_key = PeerKey(
                fqdn=connected_device.fqdn,
                addr=addr,
                vrf=getattr(connected_dto, "vrf", "")
            )
            try:
                if peer_key in connected_peers:
                    pair = merge(connected_peers[peer_key], pair)
            except MergeForbiddenError as e:
                if rule.direct_order:
                    pair_names = device.fqdn, connected_device.fqdn
                else:
                    pair_names = connected_device.fqdn, device.fqdn
                raise ValueError(
                    f"Handler `{handler_name}` provides data conflicting with "
                    f"previously loaded for device pair {pair_names} "
                    f"with addr={peer_key.addr}, vrf{peer_key.vrf}:\n" + str(e)
                ) from e
            connected_peers[peer_key] = pair

        return list(connected_peers.values())  # FIXME

    def _to_bgp_peer(self, pair: Pair, interface: Optional[str]) -> Peer:
        return to_bgp_peer(pair.local, pair.connected, pair.device.hostname, interface)

    def _virtual_to_bgp_peer(self, pair: VirtualPair, interface: Optional[str]) -> Peer:
        return to_bgp_peer(pair.local, pair.connected, "", interface)

    def _to_bgp_global(self, global_options: GlobalOptionsDTO) -> GlobalOptions:
        return to_bgp_global_options(global_options)

    def _apply_interface_changes(self, device: Device, neighbor: Device, changes: InterfaceChanges) -> str:
        port_pairs = self._storage.search_connections(device, neighbor)
        if len(port_pairs) > 1:
            if changes.lag is changes.svi is None:
                raise ValueError(
                    f"Multiple connections found between {device.fqdn} and {neighbor.fqdn}."
                    "Specify LAG or SVI"
                )
        if changes.lag is not None:
            target_interface = device.make_lag(
                lag=changes.lag,
                ports=[local_port.name for local_port, remote_port in port_pairs],
                lag_min_links=changes.lag_links_min,
            )
            if changes.subif is not None:
                target_interface = device.add_subif(target_interface.name, changes.subif)
        elif changes.subif is not None:
            # single connection
            local_port, remote_port = port_pairs[0]
            target_interface = device.add_subif(local_port.name, changes.subif)
        elif changes.svi is not None:
            target_interface = device.add_svi(changes.svi)
        else:
            target_interface, _ = port_pairs[0]
        target_interface.add_addr(changes.addr, changes.vrf)
        return target_interface.name

    def _apply_virtual_interface_changes(self, device: Device, local: VirtualLocalDTO) -> str:
        return device.add_svi(local.svi).name  # we check if SVI configured in execute method

    def execute_for(self, device: Device) -> MeshExecutionResult:
        all_fqdns = self._storage.resolve_all_fdnds()

        global_options = self._to_bgp_global(self._execute_globals(device))

        peers = []
        for direct_pair in self._execute_direct(device):
            target_interface = self._apply_interface_changes(
                device,
                direct_pair.device,
                to_interface_changes(direct_pair.local),
            )
            peers.append(self._to_bgp_peer(direct_pair, target_interface))

        for virtual_pair in self._execute_virtual(device):
            target_interface = self._apply_virtual_interface_changes(
                device,
                virtual_pair.local,
            )
            peers.append(self._virtual_to_bgp_peer(virtual_pair, target_interface))

        for connected_pair in self._execute_indirect(device, all_fqdns):
            peers.append(self._to_bgp_peer(connected_pair, None))

        return MeshExecutionResult(
            global_options=global_options,
            peers=peers,
        )



import abc
from collections.abc import Sequence
from typing import Any, Iterable, Optional, Type, Union, Protocol, Dict
from annet.connectors import Connector, get_connector_from_config
from annet.annlib.netdev.views.hardware import HardwareView


class _StorageConnector(Connector["StorageProvider"]):
    name = "Storage"  # legacy
    ep_name = "storage"  # legacy
    ep_by_group_only = "annet.connectors.storage"


storage_connector = _StorageConnector()


class StorageProvider(abc.ABC):
    @abc.abstractmethod
    def storage(self) -> Type["Storage"]:
        pass

    @abc.abstractmethod
    def opts(self) -> Type["StorageOpts"]:
        pass

    @abc.abstractmethod
    def query(self) -> Type["Query"]:
        pass

    @abc.abstractmethod
    def name(self) -> str:
        pass


class Storage(abc.ABC):
    @abc.abstractmethod
    def __enter__(self):
        pass

    @abc.abstractmethod
    def __exit__(self, _, __, ___):
        pass

    @abc.abstractmethod
    def resolve_object_ids_by_query(self, query: Any):
        pass

    @abc.abstractmethod
    def resolve_fdnds_by_query(self, query: Any):
        pass

    @abc.abstractmethod
    def resolve_all_fdnds(self) -> list[str]:
        pass

    @abc.abstractmethod
    def search_connections(self, device: "Device", neighbor: "Device") -> list[tuple["Interface", "Interface"]]:
        pass

    @abc.abstractmethod
    def make_devices(
        self,
        query: Any,
        preload_neighbors: bool = False,
        use_mesh: Optional[bool] = None,
        preload_extra_fields: bool = False,
        **kwargs,
    ):
        pass

    @abc.abstractmethod
    def get_device(self, obj_id, preload_neighbors=False, use_mesh=None, **kwargs) -> "Device":
        pass

    @abc.abstractmethod
    def flush_perf(self):
        pass


class StorageOpts(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def parse_params(cls, conf_params: Optional[Dict[str, str]], cli_opts: Any):
        pass


class Query(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def new(cls, query: Union[str, Iterable[str]], hosts_range: Optional[slice] = None) -> "Query":
        pass

    def is_empty(self) -> bool:
        return False


class Interface(Protocol):
    @property
    @abc.abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def add_addr(self, address_mask: str, vrf: Optional[str]) -> None:
        raise NotImplementedError


class Device(Protocol):
    @property
    @abc.abstractmethod
    def storage(self) -> Storage:
        pass

    @abc.abstractmethod
    def __hash__(self):
        pass

    @abc.abstractmethod
    def is_pc(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def hw(self) -> HardwareView:
        pass

    @property
    @abc.abstractmethod
    def id(self):
        pass

    @property
    @abc.abstractmethod
    def fqdn(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def hostname(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def neighbours_ids(self):
        pass

    @property
    @abc.abstractmethod
    def neighbours_fqdns(self):
        pass

    @property
    @abc.abstractmethod
    def breed(self) -> str:
        pass

    @abc.abstractmethod
    def make_lag(self, lag: int, ports: Sequence[str], lag_min_links: Optional[int]) -> Interface:
        raise NotImplementedError

    @abc.abstractmethod
    def add_svi(self, svi: int) -> Interface:
        """Add SVI interface or return existing one"""
        raise NotImplementedError

    @abc.abstractmethod
    def add_subif(self, interface: str, subif: int) -> Interface:
        """Add sub interface or return existing one"""
        raise NotImplementedError


def get_storage() -> tuple[StorageProvider, Dict[str, Any]]:
    connectors = storage_connector.get_all()
    return get_connector_from_config("storage", connectors)



"""Support JSON patch (RFC 6902) and JSON Pointer (RFC 6901) with globs."""

import copy
import fnmatch
import json
from collections.abc import Mapping, Sequence
from operator import itemgetter
from typing import Any, Dict, List, Optional

import jsonpatch
import jsonpointer


def format_json(data: Any, stable: bool = False) -> str:
    """Serialize to json."""
    return json.dumps(data, indent=4, ensure_ascii=False, sort_keys=not stable) + "\n"


def apply_json_fragment(
        old: Dict[str, Any],
        new_fragment: Dict[str, Any],
        acl: List[str],
) -> Dict[str, Any]:
    """
    Replace parts of the old document with 'new_fragment' using ACL restrictions.
    """
    full_new_config = copy.deepcopy(old)
    for acl_item in acl:
        new_pointers = _resolve_json_pointers(acl_item, new_fragment)
        old_pointers = _resolve_json_pointers(acl_item, full_new_config)

        for pointer in new_pointers:
            new_value = pointer.get(new_fragment)
            _ensure_pointer_exists(full_new_config, pointer)
            pointer.set(full_new_config, new_value)

        # delete matched parts in old config whicn are not present in the new
        paths = {p.path for p in new_pointers}
        to_delete = [p for p in old_pointers if p.path not in paths]
        for pointer in to_delete:
            doc, part = pointer.to_last(full_new_config)
            if isinstance(doc, dict) and isinstance(part, str):
                doc.pop(part, None)

    return full_new_config


def _ensure_pointer_exists(doc: Dict[str, Any], pointer: jsonpointer.JsonPointer) -> None:
    """
    Ensure that document has all pointer parts (if possible).

    This is workaround for errors of type:

    ```
    jsonpointer.JsonPointerException: member 'MY_PART' not found in {}
    ```

    See for details: https://github.com/stefankoegl/python-json-pointer/issues/41
    """
    parts_except_the_last = pointer.get_parts()[:-1]
    doc_pointer: Dict[str, Any] = doc
    for part in parts_except_the_last:
        if isinstance(doc_pointer, dict):
            if part not in doc_pointer or doc_pointer[part] is None:
                # create an empty object by the pointer part
                doc_pointer[part] = {}

            # follow the pointer to delve deeper
            doc_pointer = doc_pointer[part]
        else:
            # not a dict - cannot delve deeper
            break


def make_patch(old: Dict[str, Any], new: Dict[str, Any]) -> List[Dict[str, Any]]:
    """Generate a JSON patch by comparing the old document with the new one."""
    return sorted(jsonpatch.make_patch(old, new).patch, key=itemgetter("path"))


def apply_patch(content: Optional[bytes], patch_bytes: bytes) -> bytes:
    """
    Apply JSON patch to file contents.

    If content is None it is considered that the file does not exist.
    """
    old_doc: Any
    if content is not None:
        old_doc = json.loads(content)
    else:
        old_doc = None

    patch_data = json.loads(patch_bytes)
    patch = jsonpatch.JsonPatch(patch_data)
    new_doc = patch.apply(old_doc)

    new_contents = format_json(new_doc, stable=True).encode()
    return new_contents


def apply_acl_filters(content: Dict[str, Any], filters: List[str]) -> Dict[str, Any]:
    result = {}
    for f in filters:
        filter_text = f.strip()
        if not filter_text:
            continue

        pointers = _resolve_json_pointers(filter_text, content)
        for pointer in pointers:
            part = pointer.get(copy.deepcopy(content))

            sub_tree = result
            for i in pointer.get_parts():
                if i not in sub_tree:
                    sub_tree[i] = {}
                sub_tree = sub_tree[i]

            patch = jsonpatch.JsonPatch([{"op": "add", "path": pointer.path, "value": part}])
            result = patch.apply(result)

    return result


def _resolve_json_pointers(pattern: str, content: Dict[str, Any]) -> List[jsonpointer.JsonPointer]:
    """
    Resolve globbed json pointer pattern to a list of actual pointers, existing in the document.

    For example, given the following document:

    {
        "foo": {
            "bar": {
                "baz": [1, 2]
            },
            "qux": {
                "baz": [3, 4]
            },
        }
    }

    Pattern "/f*/*/baz" will resolve to:

    [
        "/foo/bar/baz"",
        "/foo/qux/baz",
    ]

    Pattern "/f*/q*/baz/*" will resolve to:

    [
        "/foo/qux/baz/0",
        "/foo/qux/baz/1",
    ]

    Pattern "/*" will resolve to:

    [
        "/foo"
    ]
    """
    parts = jsonpointer.JsonPointer(pattern).parts
    matched = [([], content)]
    for part in parts:
        new_matched = []
        for matched_parts, doc in matched:
            keys_and_docs = []
            if isinstance(doc, Mapping):
                keys_and_docs = [
                    (key, doc[key]) for key in doc.keys()
                    if fnmatch.fnmatchcase(key, part)
                ]
            elif isinstance(doc, Sequence):
                keys_and_docs = [
                    (str(i), doc[i]) for i in range(len(doc))
                    if fnmatch.fnmatchcase(str(i), part)
                ]
            for key, sub_doc in keys_and_docs:
                new_matched.append((matched_parts + [key], sub_doc))
        matched = new_matched

    ret: List[jsonpointer.JsonPointer] = []
    for matched_parts, _ in matched:
        ret.append(jsonpointer.JsonPointer("/" + "/".join(matched_parts)))
    return ret



import dataclasses
import itertools
import re
from collections import OrderedDict as odict
from typing import TYPE_CHECKING, Any, Dict, Iterable, Optional, Tuple, Union, List

from .types import Op

if TYPE_CHECKING:
    from .patching import PatchTree


# =====
class ParserError(Exception):
    pass


# =====
class _CommentOrEmpty:
    pass


class BlockBegin:
    pass


class BlockEnd:
    pass


RowWithContext = Tuple[str, Optional[Dict[str, Any]]]


def block_wrapper(value: Any) -> Iterable[Any]:
    yield from iter((BlockBegin, value, BlockEnd))


@dataclasses.dataclass
class FormatterContext:
    parent: Optional["FormatterContext"] = None

    prev: Optional[RowWithContext] = None
    current: Optional[RowWithContext] = None
    next: Optional[RowWithContext] = None

    @property
    def level(self) -> int:
        if self.parent is None:
            return 0
        return self.parent.level + 1

    @property
    def row_prev(self) -> Optional[str]:
        return self.prev and self.prev[0]

    @property
    def row(self) -> Optional[str]:
        return self.current and self.current[0]

    @property
    def row_next(self) -> Optional[str]:
        return self.next and self.next[0]


# =====
class CommonFormatter:
    def __init__(self, indent="  "):
        self._indent = indent
        self._block_begin = ""
        self._block_end = ""
        self._statement_end = ""

    def split(self, text):
        return list(filter(None, text.split("\n")))

    def join(self, config):
        return "\n".join(
            _filtered_block_marks(
                self._indent_blocks(self._blocks(config, is_patch=False))
            )
        )

    def diff_generator(self, diff):
        yield from self._diff_lines(diff)

    def diff(self, diff):
        return list(self.diff_generator(diff))

    def patch(self, patch):
        return "\n".join(
            _filtered_block_marks(
                self._indent_blocks(self._blocks(patch, is_patch=True))
            )
        )

    def cmd_paths(self, patch):
        ret = odict()
        path = []
        for row, context in self.blocks_and_context(patch, is_patch=True):
            if row is BlockBegin:
                path.append(path[-1])
            elif row is BlockEnd:
                path.pop()
            else:
                if path:
                    path.pop()
                path.append(row)
                ret[tuple(path)] = context
        return ret

    def patch_plain(self, patch):
        return list(self.cmd_paths(patch).keys())

    def _diff_lines(self, diff, _level=0, _block_sign=None):
        sign_map = {
            Op.REMOVED: "-",
            Op.ADDED: "+",
            Op.MOVED: ">",
            Op.AFFECTED: " ",
        }
        for (flag, row, children, _) in diff:
            sign = sign_map[flag]
            if not children:
                yield "%s %s%s" % (sign, self._indent * _level, row + self._statement_end)
            else:
                yield "%s %s%s" % (sign, self._indent * _level, row + self._block_begin)
                yield from self._diff_lines(children, _level + 1, sign)
        if _level > 0 and self._block_end and _block_sign is not None:
            yield "%s %s%s" % (_block_sign, self._indent * (_level - 1), self._block_end)

    def _indented_blocks(self, tree):
        return self._indent_blocks(self._blocks(tree, False))

    def _indent_blocks(self, blocks):
        _level = 0
        for row in blocks:
            if row is BlockBegin:
                _level += 1
            elif row is BlockEnd:
                _level -= 1
            else:
                row = self._indent * _level + row
            yield row

    def blocks_and_context(
        self,
        tree: "PatchTree",
        is_patch: bool,
        context: Optional[FormatterContext] = None
    ):
        if context is None:
            context = FormatterContext()

        if is_patch:
            items = [(item.row, item.child, item.context) for item in tree.itms]
        else:
            items = [(row, child, {}) for row, child in tree.items()]

        n = len(items)
        for i in range(n):
            prev_row, prev_sub_config, prev_row_context = items[i - 1] if i > 0 else (None, None, None)
            row, sub_config, row_context = items[i]
            next_row, next_sub_config, next_row_context = items[i + 1] if i + 1 < n else (None, None, None)

            context.current = (row, row_context)
            context.prev = (prev_row, prev_row_context) if prev_row else None
            context.next = (next_row, next_row_context) if next_row else None

            yield row, row_context

            if sub_config or (is_patch and sub_config is not None):
                yield BlockBegin, None
                yield from self.blocks_and_context(
                    sub_config, is_patch, context=FormatterContext(parent=context)
                )
                yield BlockEnd, None

    def _blocks(self, tree, is_patch):
        for row, _context in self.blocks_and_context(tree, is_patch):
            yield row


class BlockExitFormatter(CommonFormatter):
    def __init__(self, block_exit, no_block_exit=(), indent="  "):
        super().__init__(indent)
        self._block_exit = block_exit
        self._no_block_exit = tuple(no_block_exit)

    def split_remove_spaces(self, text):
        # эта регулярка заменяет 2 и более пробела на один, но оставляет пробелы в начале линии
        text = re.sub(r"(?<=\S)\ {2,}(?=\S)", " ", text)
        res = super().split(text)
        return res

    def block_exit(self, context: Optional[FormatterContext]) -> Iterable[Any]:
        current = context and context.row
        if current and not current.startswith(self._no_block_exit):
            yield from block_wrapper(self._block_exit)

    def blocks_and_context(self, tree, is_patch, context: Optional[FormatterContext] = None):
        if context is None:
            context = FormatterContext()

        level = context.level
        block_level = level

        last_row_context = {}
        for row, row_context in super().blocks_and_context(tree, is_patch, context=context):
            yield row, row_context
            if row_context is not None:
                last_row_context = row_context

            if row is BlockBegin:
                block_level += 1
            elif row is BlockEnd:
                block_level -= 1

            if row is BlockEnd and block_level == level and is_patch:
                for exit_statement in filter(None, self.block_exit(context)):
                    yield exit_statement, last_row_context


class HuaweiFormatter(BlockExitFormatter):
    def __init__(self, indent="  "):
        super().__init__(
            block_exit="quit",
            no_block_exit=[
                "rsa peer-public-key",
                "dsa peer-public-key",
                "public-key-code begin",
            ],
            indent=indent,
        )

    def split(self, text):
        # на старых прошивка наблюдается баг с двумя пробелами в этом месте в конфиге
        # например на VRP V100R006C00SPC500 + V100R006SPH003
        policy_end_blocks = ("end-list", "endif", "end-filter")
        tree = self.split_remove_spaces(text)
        tree[:] = filter(lambda x: not str(x).strip().startswith(policy_end_blocks), tree)
        return tree

    def block_exit(self, context: Optional[FormatterContext]):
        row = context and context.row or ""
        row_next = context and context.row_next
        parent_row = context and context.parent and context.parent.row or ""

        if row.startswith("xpl route-filter"):
            yield from block_wrapper("end-filter")
            return

        if row.startswith("xpl"):
            yield from block_wrapper("end-list")
            return

        if parent_row.startswith("xpl route-filter"):
            if (row.startswith(("if", "elseif")) and row.endswith("then")) and not row_next:
                yield "endif"
            elif row == "else":
                yield "endif"
            return

        yield from super().block_exit(context)


class OptixtransFormatter(CommonFormatter):
    pass


class CiscoFormatter(BlockExitFormatter):
    def __init__(self, indent="  "):
        super().__init__("exit", indent)

    def _split_indent(
        self, line: str, indent: int, block_exit_strings: List[str]
    ) -> Tuple[List[str], int]:
        """
        The small helper calculates indent shift based on block exit string.
        If configuration line has non-default block exit string it means that
        new subsection is started and indent should be increased.
        If configuration line exists in list of block exit strings,
        it means that subsection is finished and indent should be decreased

        Args:
            line: just configuration line
            indent: current indent
            block_exit_strings: list of previously seen block exit strings
        Returns:
            new indent and list of previously seen block exit strings
        """

        if line.strip() in block_exit_strings:
            indent -= 1
            block_exit_strings.remove(line.strip())
            return block_exit_strings, indent

        block_exit_wrapped = [
            v for v in self.block_exit(FormatterContext(current=(line.strip(), {})))
        ]

        if not block_exit_wrapped or len(block_exit_wrapped) != 3:
            return block_exit_strings, indent
        if not isinstance(block_exit_wrapped[1], str):
            return block_exit_strings, indent
        if block_exit_wrapped[1] == self._block_exit:
            return block_exit_strings, indent

        indent += 1
        block_exit_strings.append(block_exit_wrapped[1])
        return block_exit_strings, indent

    def split(self, text):
        additional_indent = 0
        block_exit_strings = [self._block_exit]
        tree = self.split_remove_spaces(text)
        for i, item in enumerate(tree):
            block_exit_strings, new_indent = self._split_indent(
                item, additional_indent, block_exit_strings
            )
            tree[i] = f"{' ' * additional_indent}{item}"
            additional_indent = new_indent
        return tree

    def block_exit(self, context: Optional[FormatterContext]) -> str:
        current = context and context.row or ""

        if current.startswith(("address-family")):
            yield from block_wrapper("exit-address-family")
        else:
            yield from super().block_exit(context)


class NexusFormatter(BlockExitFormatter):
    def __init__(self, indent="  "):
        super().__init__("exit", indent)

    def split(self, text):
        return self.split_remove_spaces(text)


class B4comFormatter(BlockExitFormatter):
    def __init__(self, indent="  "):
        super().__init__("exit", indent)

    def split(self, text):
        return self.split_remove_spaces(text)


class ArubaFormatter(BlockExitFormatter):
    def __init__(self, indent="  "):
        super().__init__("exit", indent)

    def split(self, text):
        return self.split_remove_spaces(text)


class AristaFormatter(BlockExitFormatter):
    def __init__(self, indent="  "):
        super().__init__("exit", indent)

    def split(self, text):
        return self.split_remove_spaces(text)


class AsrFormatter(BlockExitFormatter):
    def __init__(self, indent="  "):
        super().__init__("exit", indent)

    def split(self, text):
        policy_end_blocks = ("end-set", "endif", "end-policy")
        tree = self.split_remove_spaces(text)
        tree[:] = filter(lambda x: not x.endswith(policy_end_blocks), tree)
        return tree

    def block_exit(self, context: Optional[FormatterContext]) -> str:
        current = context and context.row or ""

        if current.startswith(("prefix-set", "as-path-set", "community-set")):
            yield from block_wrapper("end-set")
        elif current.startswith("if") and current.endswith("then"):
            yield from block_wrapper("endif")
        elif current.startswith("route-policy"):
            yield from block_wrapper("end-policy")
        else:
            yield from super().block_exit(context)


class JuniperFormatter(CommonFormatter):
    patch_set_prefix = "set "

    def __init__(self, indent="    "):
        super().__init__(indent)
        self._block_begin = " {"
        self._block_end = "}"
        self._statement_end = ";"
        self._endofline_comment = "; ##"

    def split(self, text):
        sub_regexs = (
            (re.compile(self._block_begin + r"\s*" + self._block_end + r"$"), ""),  # collapse empty blocks
            (re.compile(self._block_begin + "(\t# .+)?$"), ""),
            (re.compile(self._statement_end + r"$"), ""),
            (re.compile(r"\s*" + self._block_end + "(\t# .+)?$"), ""),
            (re.compile(self._endofline_comment + r".*$"), ""),
        )
        split = []
        for line in text.split("\n"):
            for (regex, repl_line) in sub_regexs:
                line = regex.sub(repl_line, line)
            split.append(line)
        return list(filter(None, split))

    def join(self, config):
        return "\n".join(_filtered_block_marks(self._formatted_blocks(self._indented_blocks(config))))

    def patch(self, patch):
        return "\n".join(" ".join(x) for x in self.cmd_paths(patch))

    def patch_plain(self, patch):
        return list(self.cmd_paths(patch).keys())

    def _formatted_blocks(self, blocks):
        level = 0
        line = None
        for new_line in blocks:
            if new_line is BlockBegin:
                level += 1
                if isinstance(line, str):
                    yield line + self._block_begin
            elif new_line is BlockEnd:
                level -= 1
                if isinstance(line, str):
                    yield line + self._statement_end
                yield self._indent * level + self._block_end
            elif isinstance(line, str):
                yield line + self._statement_end
            line = new_line
        if isinstance(line, str):
            yield line + self._statement_end

    def cmd_paths(self, patch, _prev=""):
        commands = odict()
        for item in patch.itms:
            key, childs, context = item.row, item.child, item.context
            if childs:
                for k, v in self.cmd_paths(childs, _prev + " " + key).items():
                    commands[k] = v
            else:
                if key.startswith("delete"):
                    cmd = "delete" + _prev + " " + key.replace("delete", "", 1).strip()
                elif key.startswith("activate"):
                    cmd = "activate" + _prev + " " + key.replace("activate", "", 1).strip()
                elif key.startswith("deactivate"):
                    cmd = "deactivate" + _prev + " " + key.replace("deactivate", "", 1).strip()
                else:
                    cmd = (self.patch_set_prefix + _prev.strip()).strip() + " " + key
                # Expanding [ a b c ] junipers list of arguments
                matches = re.search(r"^(.*)\s+\[(.+)\]$", cmd)
                if matches:
                    for c in matches.group(2).split(" "):
                        if c.strip():
                            cmd = " ".join([matches.group(1), c])
                            commands[(cmd,)] = context
                else:
                    commands[(cmd,)] = context

        return commands


class RibbonFormatter(JuniperFormatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._endofline_comment = "; # SECRET-DATA"


class JuniperList:
    """
    Форматирует inline-листы в конфиге juniper
    """

    def __init__(self, *args, spaces=True, **kwargs):
        self._items = list(*args, **kwargs)
        self.spaces = spaces

    def __str__(self):
        if self.spaces:
            return "[ %s ]" % " ".join(str(_) for _ in self._items)
        else:
            return "[%s]" % " ".join(str(_) for _ in self._items)


class NokiaFormatter(JuniperFormatter):
    patch_set_prefix = "/configure "

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._statement_end = ""
        self._endofline_comment = " ##"

    def split(self, text):
        ret = super().split(text)
        # NOCDEVDUTY-248 сдергиваем верхний configure-блок
        # NOCDEVDUTY-282 после configure {} блока могут идти еще блоки которые нам не нужны
        start, finish = None, None
        for i, line in enumerate(ret):
            if line.startswith("#"):
                continue
            # начало configure-блока
            if line == "configure":
                start = i + 1
            # любой после configure последующий блок на глобальном уровне
            elif len(line) == len(line.lstrip()):
                if start is not None and finish is None:
                    finish = i
        # Если configure-блока не было - то весь конфиг считаем configre'ом
        start = start if start is not None else 0
        finish = finish if finish is not None else len(ret)
        return ret[start:finish]

    def cmd_paths(self, patch, _prev=""):
        commands = odict()
        for item in patch.itms:
            key, childs, context = item.row, item.child, item.context
            if childs:
                for k, v in self.cmd_paths(childs, _prev + " " + key).items():
                    commands[k] = v
            else:
                if key.startswith("delete"):
                    cmd = "/configure delete" + _prev + " " + key.replace("delete", "", 1).strip()
                else:
                    cmd = self.patch_set_prefix + _prev.strip() + " " + key
                # Expanding [ a b c ] junipers list of arguments
                matches = re.search(r"^(.*)\s+\[(.+)\]$", cmd)
                if matches:
                    for c in matches.group(2).split(" "):
                        if c.strip():
                            cmd = " ".join([matches.group(1), c])
                            commands[(cmd,)] = context
                else:
                    commands[(cmd,)] = context
        return commands


class RosFormatter(CommonFormatter):
    patch_set_prefix = "set "

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._block_begin = "/"

    def join(self, config):
        return "\n".join(_filtered_block_marks(self._formatted_blocks(self._indented_blocks(config))))

    def patch(self, patch):
        return "\n".join(" ".join(x) for x in self.cmd_paths(patch))

    def patch_plain(self, patch):
        return list(self.cmd_paths(patch).keys())

    def blocks_and_context(
        self,
        tree: "PatchTree",
        is_patch: bool,
        context: Optional[FormatterContext] = None
    ):
        rows = []

        if is_patch:
            items = ((item.row, item.child, item.context) for item in tree.itms)
        else:
            items = ((row, child, {}) for row, child in tree.items())

        for row, sub_config, row_context in items:
            if sub_config or (is_patch and sub_config is not None):
                rows.append((row, sub_config, row_context))
            else:
                rows.append((row, None, row_context))

        prev_prow = None
        prev_prow_context = {}
        for sub_config, row_group in itertools.groupby(rows, lambda x: x[1]):
            if sub_config is None:
                if prev_prow:
                    yield prev_prow, prev_prow_context
                    yield BlockBegin, None
                for row, _, row_context in row_group:
                    yield row, row_context
                if prev_prow:
                    yield BlockEnd, None
            else:
                for row, _, row_context in row_group:
                    if context and context.parent and context.parent.row:
                        prev_prow, prev_prow_context = context.parent.current
                        prow = f"{context.parent.row} {row}"
                    else:
                        prow = row
                    yield prow, row_context

                    yield BlockBegin, None
                    yield from self.blocks_and_context(
                        sub_config,
                        is_patch,
                        context=FormatterContext(parent=context, current=(prow, row_context))
                    )
                    yield BlockEnd, None

    def _formatted_blocks(self, blocks):
        line = None
        for new_line in blocks:
            if new_line is BlockBegin:
                if isinstance(line, str):
                    yield self._block_begin + line.strip()
            elif isinstance(line, str):
                yield line
            line = new_line

    def _splitter_file(self, lines):
        filedesrc_re = re.compile(r"^\s+(?P<num>\d+)\s+name=\"(?P<name>[^\"]+)\"\s+type=\"(?P<type>[^\"]+)\""
                                  r"\s+(size=(?P<size>.*))?creation-time=(?P<time>.*?)(contents=(?P<content>.*)?)?$")
        file_content_indent = re.compile(r"^\s{5}")
        out = []
        files = {}
        curfile = None
        for line in lines:
            match = filedesrc_re.search(line)
            if match:
                if match.group("type").strip() == ".txt file":
                    curfile = match.group("name")
                    files[curfile] = {"name": curfile, "contents": []}
                    if match.group("content"):
                        files[curfile]["contents"].append(match.group("content").strip())
            elif curfile and file_content_indent.match(line):
                files[curfile]["contents"].append(file_content_indent.sub("", line))
        for file in files.values():
            out.append(f"print file={file['name']}")
            if len(file["contents"]) > 0:
                text = "\n".join(file["contents"])
                out.append(f"set {file['name']} contents=\"{text}\"")
        return out

    def _splitter_user_ssh_keys(self, lines):
        keydescr_re = re.compile(r"user=(?P<user>\w+).*key-owner=(?P<owner>.*)$")
        out = []
        for line in lines:
            match = keydescr_re.search(line)
            if match:
                out.append(f"import public-key-file={match.group('owner')}.ssh_key.txt user={match.group('user')}")

        return out

    def split(self, text):
        split = []
        level = 0
        postj = {}
        curgroup = None
        for line in text.split("\n"):
            if line.startswith("/"):
                if curgroup:
                    for row in getattr(self, curgroup)(postj[curgroup]):
                        if level > 0:
                            row = row.strip()
                        split.append(self._indent * level + row)

                level = 0
                for group in line.split():
                    split.append(self._indent * level + group.replace("/", ""))
                    level += 1

                gpath = line.replace("/", "_splitter_").replace(" ", "_").replace("-", "_")
                if hasattr(self, gpath):
                    postj[gpath] = []
                    curgroup = gpath
                else:
                    curgroup = None
            else:
                row = line
                if curgroup:
                    postj[curgroup].append(row)
                else:
                    if level > 0:
                        row = line.strip()
                    split.append(self._indent * level + row)
        if curgroup:
            for row in getattr(self, curgroup)(postj[curgroup]):
                if level > 0:
                    row = row.strip()
                split.append(self._indent * level + row)
        return list(filter(None, split))

    def cmd_paths(self, patch, _prev=""):
        rm_regexs = (
            (re.compile(r"^add "), ""),
            (re.compile(r"^print file="), "name="),
        )
        patch_items = []
        for item in patch.itms:
            key, childs, context = item.row, item.child, item.context
            if childs:
                patch_items.append((key, childs, context))
            else:
                patch_items.append((key, None, context))

        commands = odict()
        prev_cmd = None
        prev_context = None
        for childs, items in itertools.groupby(patch_items, lambda x: x[1]):
            if childs is None:
                if prev_cmd:
                    commands[(prev_cmd,)] = prev_context
                for key, _, context in items:
                    if key.startswith("remove"):
                        find_cmd = key.replace("remove", "", 1).strip()
                        for (regex, repl_line) in rm_regexs:
                            find_cmd = regex.sub(repl_line, find_cmd)
                        cmd = "remove [ find " + find_cmd + " ]"
                    else:
                        cmd = key
                    commands[(cmd,)] = context
            else:
                for key, _, context in items:
                    if _prev:
                        prev_cmd = _prev
                        prev_context = context
                        block_cmd = f"{_prev} {key}"
                    else:
                        block_cmd = f"/{key}"
                    commands[(block_cmd,)] = context
                    for k, v in self.cmd_paths(childs, block_cmd).items():
                        commands[k] = v
        return commands


def make_formatter(vendor, **kwargs):
    formatters = {
        "juniper": JuniperFormatter,
        "cisco": CiscoFormatter,
        "nexus": NexusFormatter,
        "huawei": HuaweiFormatter,
        "optixtrans": OptixtransFormatter,
        "arista": AristaFormatter,
        "nokia": NokiaFormatter,
        "routeros": RosFormatter,
        "aruba": ArubaFormatter,
        "pc": CommonFormatter,
        "ribbon": RibbonFormatter,
        "b4com": B4comFormatter,
    }
    return formatters[vendor](**kwargs)


# ====
def parse_to_tree(text, splitter, comments=("!", "#")):
    tree = odict()
    for stack in _stacked(splitter(text), tuple(comments)):
        local_tree = tree
        for key in stack:
            if key not in local_tree:
                local_tree[key] = odict()
            local_tree = local_tree[key]
    return tree


# =====
def _stacked(lines, comments):
    stack = []
    for (level, line) in _stripped_indents(lines, comments):
        level += 1
        if level > len(stack):
            stack.append(line)
        elif level == len(stack):
            stack[-1] = line
        else:
            stack = stack[:level - 1] + [line]
        yield tuple(stack)


def _stripped_indents(lines, comments):
    indents = []
    curr_level = 0
    g_level = None

    for (number, (level, line)) in enumerate(_parsed_indents(lines, comments), start=1):
        if isinstance(line, str):
            if g_level is None:
                g_level = level
            level = level - (g_level or 0)
            if level < 0:
                raise ParserError("Invalid top indention: line %d: %s" % (number, line))

            if level > curr_level:
                indents.append(level - curr_level)
                curr_level += level - curr_level
            elif level < curr_level:
                while curr_level > level and len(indents):
                    curr_level -= indents.pop()
                if curr_level != level:
                    raise ParserError("Invalid top indention: line %d: %s" % (number, line))

            yield (len(indents), line)

        elif line is BlockEnd:
            indents = []
            curr_level = 0
            g_level = None


def _parsed_indents(lines, comments):
    for line in _filtered_lines(lines, comments):
        if isinstance(line, str):
            yield (_parse_indent(line), line.strip())
        else:
            yield (0, line)


def _filtered_lines(lines, comments):
    for line in lines:
        stripped = line.strip()
        # TODO Это для хуавей, так что хелпер нужно унести в Formatter
        if "#" in comments and line.startswith("#"):
            yield BlockEnd
        elif len(stripped) == 0 or stripped.startswith(comments):
            yield _CommentOrEmpty
        else:
            yield line


def _filtered_block_marks(blocks):
    return filter(lambda b: isinstance(b, str), blocks)


def _parse_indent(line):
    level = 0
    for ch in line:
        if ch in ("\t", " "):
            level += 1
        else:
            break
    return level



from logging import getLogger

from annet.annlib.netdev.views.hardware import HardwareView

logger = getLogger(__name__)

_VENDORS = {
    "cisco": "Cisco",
    "catalyst": "Cisco Catalyst",
    "nexus": "Cisco Nexus",
    "huawei": "Huawei",
    "optixtrans": "Huawei OptiXtrans",
    "juniper": "Juniper",
    "arista": "Arista",
    "pc": "PC",
    "nokia": "Nokia",
    "aruba": "Aruba",
    "routeros": "RouterOS",
    "ribbon": "Ribbon",
    "b4com": "B4com",
    "h3c": "H3C",
}


def get_hw(manufacturer: str, model: str, platform_name: str):
    # by some reason Netbox calls Mellanox SN as MSN, so we fix them here
    if manufacturer == "Mellanox" and model.startswith("MSN"):
        model = model.replace("MSN", "SN", 1)
    vendor = manufacturer + " " + model
    hw = HardwareView(_VENDORS.get(vendor.lower(), vendor), platform_name)
    return hw


def get_breed(manufacturer: str, model: str):
    if manufacturer == "Huawei" and model.startswith("CE"):
        return "vrp85"
    elif manufacturer == "Huawei" and model.startswith("NE"):
        return "vrp85"
    elif manufacturer == "Huawei":
        return "vrp55"
    elif manufacturer == "H3C":
        return "h3c"
    elif manufacturer in ("Mellanox", "NVIDIA"):
        return "cuml2"
    elif manufacturer == "Juniper":
        return "jun10"
    elif manufacturer == "Cisco":
        return "ios12"
    elif manufacturer == "Adva":
        return "adva8"
    elif manufacturer == "Arista":
        return "eos4"
    elif manufacturer == "B4com":
        return "bcom-os"
    elif manufacturer == "MikroTik":
        return "routeros"
    elif manufacturer == "PC":
        return "pc"
    return ""



from __future__ import annotations

import re
from typing import (
    Iterable,
    List,
    Set,
    Union,
)

from annet.lib import (
    add_annotation,
    flatten,
)
from .base import TreeGenerator, _filter_str
from .exceptions import NotSupportedDevice


class PartialGenerator(TreeGenerator):
    TYPE = "PARTIAL"
    TAGS: List[str] = []

    def __init__(self, storage):
        super().__init__()
        self.storage = storage
        self._annotate = False
        self._running_gen = None
        self._annotations = []
        self._annotation_module = self.__class__.__module__ or ""

    def supports_device(self, device) -> bool:
        if self.__class__.run is PartialGenerator.run:
            return bool(self._get_vendor_func(device.hw.vendor, "run"))
        else:
            return True

    def acl(self, device):
        acl_func = self._get_vendor_func(device.hw.vendor, "acl")
        if acl_func:
            return acl_func(device)

    def acl_safe(self, device):
        acl_func = self._get_vendor_func(device.hw.vendor, "acl_safe")
        if acl_func:
            return acl_func(device)

    def run(self, device) -> Iterable[Union[str, tuple]]:
        run_func = self._get_vendor_func(device.hw.vendor, "run")
        if run_func:
            return run_func(device)

    def get_user_runner(self, device):
        if self.__class__.run is not PartialGenerator.run:
            return self.run
        return self._get_vendor_func(device.hw.vendor, "run")

    def _get_vendor_func(self, vendor: str, func_name: str):
        attr_name = f"{func_name}_{vendor}"
        if hasattr(self, attr_name):
            return getattr(self, attr_name)
        return None

    # =====

    @classmethod
    def get_aliases(cls) -> Set[str]:
        return {cls.__name__, *cls.TAGS}

    def __call__(self, device, annotate=False):
        self._indents = []
        self._rows = []
        self._running_gen = self.run(device)
        self._annotate = annotate

        if annotate and self.__class__.__module__:
            self._annotation_module = ".".join(
                self.__class__.__module__.split(".")[-2:])

        for text in self._running_gen:
            if isinstance(text, tuple):
                text = " ".join(map(_filter_str, flatten(text)))
            else:
                text = _filter_str(text)
            self._append_text(text)

        for row in self._rows:
            assert re.search(r"\bNone\b", row) is None, \
                "Found 'None' in yield result: %s" % (row)
        if annotate:
            generated_rows = (
                add_annotation(x, y)
                for (x, y) in zip(self._rows, self._annotations)
            )
        else:
            generated_rows = self._rows
        return "\n".join(generated_rows) + "\n"

    def _append_text(self, text):
        def annotation_cb(row):
            annotation = "%s:%d" % self.get_running_line()
            self._annotations.append(annotation)
            return row

        self._append_text_cb(
            text,
            annotation_cb if self._annotate else None
        )

    def get_running_line(self):
        if not self._running_gen or not self._running_gen.gi_frame:
            return (repr(self._running_gen), -1)
        return self._annotation_module, self._running_gen.gi_frame.f_lineno

    @classmethod
    def literal(cls, item):
        return '"{}"'.format(item)

    def __repr__(self):
        return "<%s>" % self.__class__.__name__



__all__ = [
    "MatchField",
    "ThenField",
    "RouteMap",
    "Route",
    "ResultType",
    "ActionType",
    "Action",
    "SingleAction",
    "AndCondition",
    "R",
    "ConditionOperator",
    "Condition",
    "SingleCondition",
    "RoutingPolicyStatement",
    "RoutingPolicy",
    "CommunityActionValue",
    "PrefixMatchValue",
]

from .action import Action, ActionType, SingleAction
from .condition import AndCondition, Condition, ConditionOperator, SingleCondition
from .match_builder import R, MatchField, PrefixMatchValue
from .policy import RoutingPolicyStatement, RoutingPolicy
from .result import ResultType
from .routemap import RouteMap, Route
from .statement_builder import ThenField, CommunityActionValue



from dataclasses import dataclass
from ipaddress import ip_interface
from typing import Optional, Union

from adaptix import Retort, loader, Chain, name_mapping, as_is_loader

from .peer_models import DirectPeerDTO, IndirectPeerDTO, VirtualPeerDTO, VirtualLocalDTO
from ..bgp_models import (
    Aggregate, GlobalOptions, VrfOptions, FamilyOptions, Peer, PeerGroup, ASN, PeerOptions,
    Redistribute, BFDTimers,
)


PeerDTO = Union[DirectPeerDTO, IndirectPeerDTO, VirtualPeerDTO]
LocalDTO = Union[DirectPeerDTO, IndirectPeerDTO, VirtualLocalDTO]


@dataclass
class InterfaceChanges:
    addr: str
    lag: Optional[int] = None
    lag_links_min: Optional[int] = None
    svi: Optional[int] = None
    subif: Optional[int] = None
    vrf: Optional[str] = None

    def __post_init__(self):
        if self.lag is not None and self.svi is not None:
            raise ValueError("Cannot use LAG and SVI together")
        if self.svi is not None and self.subif is not None:
            raise ValueError("Cannot use Subif and SVI together")


class ObjMapping:
    def __init__(self, obj):
        self.obj = obj

    def __contains__(self, item):
        return hasattr(self.obj, item)

    def __getitem__(self, name):
        return getattr(self.obj, name)

    def get(self, name, default=None):
        return getattr(self.obj, name, default)


retort = Retort(
    recipe=[
        loader(InterfaceChanges, ObjMapping, Chain.FIRST),
        loader(ASN, ASN),
        loader(GlobalOptions, ObjMapping, Chain.FIRST),
        loader(VrfOptions, ObjMapping, Chain.FIRST),
        loader(FamilyOptions, ObjMapping, Chain.FIRST),
        loader(Aggregate, ObjMapping, Chain.FIRST),
        loader(PeerOptions, ObjMapping, Chain.FIRST),
        as_is_loader(Redistribute),
        as_is_loader(BFDTimers),
        name_mapping(PeerOptions, map={
            "local_as": "asnum",
        }),
        loader(list[PeerGroup], lambda x: list(x.values()), Chain.FIRST),
        loader(PeerGroup, ObjMapping, Chain.FIRST),
    ]
)

to_bgp_global_options = retort.get_loader(GlobalOptions)
to_interface_changes = retort.get_loader(InterfaceChanges)


def to_bgp_peer(local: LocalDTO, connected: PeerDTO, connected_hostname: str, interface: Optional[str]) -> Peer:
    options = retort.load(local, PeerOptions)
    result = Peer(
        addr=str(ip_interface(connected.addr).ip),
        interface=interface,
        remote_as=ASN(connected.asnum),
        hostname=connected_hostname,
        options=options,
    )
    # connected
    result.vrf_name = getattr(connected, "vrf", result.vrf_name)
    result.group_name = getattr(connected, "group_name", result.group_name)
    result.description = getattr(connected, "description", result.description)
    result.families = getattr(connected, "families", result.families)
    # local
    result.import_policy = getattr(local, "import_policy", result.import_policy)
    result.export_policy = getattr(local, "export_policy", result.export_policy)
    result.update_source = getattr(local, "update_source", result.update_source)
    return result



from collections.abc import Iterator, Sequence
from typing import Any, cast

from annet.generators import PartialGenerator, BaseGenerator
from annet.rpl import (
    CommunityActionValue,
    ResultType, RoutingPolicyStatement, RoutingPolicy, ConditionOperator, SingleCondition, SingleAction, ActionType,
)
from annet.rpl.statement_builder import AsPathActionValue, NextHopActionValue
from annet.storage import Storage
from .items import AS_PATH_FILTERS, COMMUNITIES, EXT_COMMUNITIES
from .route_policy import routemap

HUAWEI_MATCH_COMMAND_MAP = {
    "as_path_filter": "as-path-filter {option_value}",
    "metric": "cost {option_value}",
    "protocol": "protocol {option_value}",
    "interface": "interface {option_value}",
}

HUAWEI_THEN_COMMAND_MAP = {
    "metric": "cost {option_value}",
    "local_pref": "local-preference {option_value}",
    "metric_type": "cost-type {option_value}",
    "mpls_label": "mpls-label",
    "origin": "origin {option_value}",
    "tag": "tag {option_value}",
    # unsupported: resolution
    # unsupported: rpki_valid_state
}
HUAWEI_RESULT_MAP = {
    ResultType.ALLOW: "permit",
    ResultType.DENY: "deny",
    ResultType.NEXT: ""
}


class RoutingPolicyGenerator(PartialGenerator):
    TAGS = ["policy", "rpl", "routing"]

    def acl_huawei(self, _):
        return r"""
        ip as-path-filter
        route-policy *
            ~ %global=1
        """

    def _huawei_match(self, device, condition: SingleCondition[Any]) -> Iterator[Sequence[str]]:
        if condition.field == "community":
            if condition.operator is ConditionOperator.HAS:
                if len(condition.value) > 1:
                    raise NotImplementedError("Multiple HAS for communities is not supported for huawei")
            elif condition.operator is not ConditionOperator.HAS_ANY:
                raise NotImplementedError("Community operator %r not supported for huawei" % condition.operator)
            for comm_name in condition.value:
                yield "if-match community-filter", comm_name
            return
        if condition.field == "extcommunity":
            if condition.operator is ConditionOperator.HAS:
                if len(condition.value) > 1:
                    raise NotImplementedError("Multiple HAS for extcommunities is not supported for huawei")
            elif condition.operator is not ConditionOperator.HAS_ANY:
                raise NotImplementedError("Extcommunity operator %r not supported for huawei" % condition.operator)
            for comm_name in condition.value:
                yield "if-match extcommunity-filter", comm_name
            return
        if condition.field == "ip_prefix":
            for name in condition.value.names:
                yield "if-match", "ip-prefix-filter", name
            return
        if condition.field == "ipv6_prefix":
            for name in condition.value.names:
                yield "if-match", "ipv6 address prefix-list", name
            return
        if condition.field == "as_path_length":
            if condition.operator is ConditionOperator.EQ:
                yield "if-match", "as-path length", condition.value
            elif condition.operator is ConditionOperator.LE:
                yield "if-match", "as-path length less-equal", condition.value
            elif condition.operator is ConditionOperator.GE:
                yield "if-match", "as-path length greater-equal", condition.value
            elif condition.operator is ConditionOperator.BETWEEN_INCLUDED:
                yield "if-match", "as-path length greater-equal", condition.value[0], "less-equal", condition.value[1]
            else:
                raise NotImplementedError(
                    f"as_path_length operator {condition.operator} not supported for huawei",
                )
            return
        if condition.operator is not ConditionOperator.EQ:
            raise NotImplementedError(
                f"`{condition.field}` with operator {condition.operator} is not supported for huawei",
            )
        if condition.field not in HUAWEI_MATCH_COMMAND_MAP:
            raise NotImplementedError(f"Match using `{condition.field}` is not supported for huawei")
        cmd = HUAWEI_MATCH_COMMAND_MAP[condition.field]
        yield "if-match", cmd.format(option_value=condition.value)

    def _huawei_then_community(self, action: SingleAction[CommunityActionValue]) -> Iterator[Sequence[str]]:
        if action.value.replaced is not None:
            if not action.value.replaced:
                yield "apply", "community", "none"
            first = True
            for community_name in action.value.replaced:
                community = COMMUNITIES[community_name]
                for comm_value in community.values:
                    if first:
                        yield "apply", "community", comm_value
                        first = False
                    else:
                        yield "apply", "community", comm_value, "additive"
        for community_name in action.value.added:
            community = COMMUNITIES[community_name]
            for comm_value in community.values:
                yield "apply", "community", comm_value, "additive"
        for community_name in action.value.removed:
            yield "apply comm-filter", community_name, "delete"

    def _huawei_then_extcommunity(self, action: SingleAction[CommunityActionValue]) -> Iterator[Sequence[str]]:
        if action.value.replaced is not None:
            if not action.value.replaced:
                yield "apply", "extcommunity", "none"
            first = True
            for community_name in action.value.replaced:
                community = EXT_COMMUNITIES[community_name]
                for comm_value in community.values:
                    if first:
                        yield "apply", "extcommunity", comm_value
                        first = False
                    else:
                        yield "apply", "extcommunity", comm_value, "additive"
        for community_name in action.value.added:
            community = EXT_COMMUNITIES[community_name]
            for comm_value in community.values:
                yield "apply", "extcommunity", comm_value, "additive"
        for community_name in action.value.removed:
            yield "apply extcommunity-filter", community_name, "delete"

    def _huawei_then(self, device, action: SingleAction[Any]) -> Iterator[Sequence[str]]:
        if action.field == "community":
            yield from self._huawei_then_community(cast(SingleAction[CommunityActionValue], action))
            return
        if action.field == "extcommunity":
            yield from self._huawei_then_extcommunity(cast(SingleAction[CommunityActionValue], action))
            return
        if action.field == "metric":
            if action.type is ActionType.ADD:
                yield "apply", f"cost + {action.value}"
            elif action.type is ActionType.SET:
                yield "apply", f"cost {action.value}"
            else:
                raise NotImplementedError(f"Action type {action.type} for metric is not supported for huawei")
            return
        if action.field == "as_path":
            as_path_action_value = cast(AsPathActionValue, action.value)
            if as_path_action_value.set is not None:
                if not as_path_action_value.set:
                    yield "apply", "as_path", "none overwrite"
                first = True
                for path_item in as_path_action_value.set:
                    if first:
                        yield "apply as-path", path_item, "overwrite"
                        first = False
                    else:
                        yield "apply as-path", path_item, "additive"
            if as_path_action_value.prepend:
                for path_item in as_path_action_value.prepend:
                    yield "apply as-path", path_item, "additive"
            if as_path_action_value.expand:  # same as prepend?
                for path_item in as_path_action_value.expand:
                    yield "apply as-path", path_item, "additive"
            if as_path_action_value.delete:
                for path_item in as_path_action_value.delete:
                    yield "apply as-path", path_item, "delete"
            if as_path_action_value.expand_last_as:
                raise RuntimeError("asp_path.expand_last_as is not supported for huawei")
            return
        if action.field == "next_hop":
            next_hop_action_value = cast(NextHopActionValue, action.value)
            if next_hop_action_value.target == "self":
                yield "apply", "cost 1"
            elif next_hop_action_value.target == "discard":
                pass
            elif next_hop_action_value.target == "peer":
                pass
            elif next_hop_action_value.target == "ipv4_addr":
                yield "apply", f"ip-address next-hop {next_hop_action_value.addr}"
            elif next_hop_action_value.target == "ipv6_addr":
                yield "apply", f"ipv6 next-hop {next_hop_action_value.addr}"
            elif next_hop_action_value.target == "mapped_ipv4":
                yield "apply", f"ipv6 next-hop ::FFFF:{next_hop_action_value.addr}"
            else:
                raise RuntimeError(f"Next_hop target {next_hop_action_value.target} is not supported for huawei")

        if action.type is not ActionType.SET:
            raise NotImplementedError(f"Action type {action.type} for `{action.field}` is not supported for huawei")
        if action.field not in HUAWEI_THEN_COMMAND_MAP:
            raise NotImplementedError(f"Then action using `{action.field}` is not supported for huawei")
        cmd = HUAWEI_THEN_COMMAND_MAP[action.field]
        yield "apply", cmd.format(option_value=action.value)

    def _huawei_statement(
            self, device, policy: RoutingPolicy, statement: RoutingPolicyStatement,
    ) -> Iterator[Sequence[str]]:
        if "as_path_filter" in statement.match:
            as_path_condition = statement.match["as_path_filter"]
            as_filter_value = AS_PATH_FILTERS[as_path_condition.value]
            yield "ip as-path-filter", \
                as_path_condition.value, \
                "index 10 permit", \
                "_{}_".format("_".join(("%s" % x for x in as_filter_value if x != ".*")))

        with self.block(
                "route-policy", policy.name,
                HUAWEI_RESULT_MAP[statement.result],
                "node", statement.number
        ):
            for condition in statement.match:
                yield from self._huawei_match(device, condition)
            for action in statement.then:
                yield from self._huawei_then(device, action)
            if statement.result is ResultType.NEXT:
                yield "goto next-node"

    def run_huawei(self, device):
        for policy in routemap.apply(device):
            for statement in policy.statements:
                yield from self._huawei_statement(device, policy, statement)


def get_generators(store: Storage) -> list[BaseGenerator]:
    return [
        RoutingPolicyGenerator(store),
    ]



from __future__ import annotations

import contextlib
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Set,
    Union,
)

from annet.storage import Device, Storage
from .base import TreeGenerator, _filter_str
from .exceptions import NotSupportedDevice


class JSONFragment(TreeGenerator):
    """Generates parts of JSON config file."""

    TYPE = "JSON_FRAGMENT"
    TAGS: List[str] = []

    def __init__(self, storage: Storage):
        super().__init__()
        self.storage = storage
        self._json_config: Dict[str, Any] = {}
        self._config_pointer: List[str] = []

        # if two generators edit same file, commands from generator with greater `reload_prio` will be used
        if not hasattr(self, "reload_prio"):
            self.reload_prio = 100

    def supports_device(self, device: Device):
        return bool(self.path(device))

    def path(self, device: Device) -> Optional[str]:
        raise NotImplementedError("Required PATH for JSON_FRAGMENT generator")

    @classmethod
    def get_aliases(cls) -> Set[str]:
        return {cls.__name__, *cls.TAGS}

    def acl(self, device: Device) -> Union[str, List[str]]:
        """
        Restrict the generator to a specified ACL using JSON Pointer syntax.

        Expected ACL to be a list of strings, but a single string is also allowed.
        """
        raise NotImplementedError("Required ACL for JSON_FRAGMENT generator")

    def acl_safe(self, device: Device) -> Union[str, List[str]]:
        """
        Restrict the generator to a specified safe ACL using JSON Pointer syntax.

        Expected ACL to be a list of strings, but a single string is also allowed.
        """
        raise NotImplementedError("Required ACL for JSON_FRAGMENT generator")

    def run(self, device: Device):
        raise NotImplementedError

    def get_reload_cmds(self, device: Device) -> str:
        ret = self.reload(device) or ""
        return ret

    def reload(self, device) -> Optional[str]:
        raise NotImplementedError

    @contextlib.contextmanager
    def block(self, *tokens, indent=None):  # pylint: disable=unused-argument
        block_str = "".join(map(_filter_str, tokens))
        self._config_pointer.append(block_str)
        try:
            yield
        finally:
            self._config_pointer.pop()

    @contextlib.contextmanager
    def block_piped(self, *tokens, indent=None):  # pylint: disable=unused-argument
        block_str = "|".join(map(_filter_str, tokens))
        self._config_pointer.append(block_str)
        try:
            yield
        finally:
            self._config_pointer.pop()

    def __call__(self, device: Device, annotate: bool = False):
        try:
            for cfg_fragment in self.run(device):
                self._set_or_replace_dict(self._config_pointer, cfg_fragment)
            return self._json_config
        finally:
            self._json_config = {}

    def _set_or_replace_dict(self, pointer, value):
        if not pointer:
            if self._json_config == {}:
                self._json_config = self.process_value(value)
            else:
                self._json_config = [self._json_config, self.process_value(value)]
        else:
            self._set_dict(self._json_config, pointer, value)

    def process_scalar_value(self, value: Any) -> Any:
        return str(value)

    def process_value(self, value: Any) -> Any:
        if isinstance(value, (list, set, frozenset)):
            return [self.process_value(x) for x in value]
        elif isinstance(value, dict):
            for k, v in value.items():
                value[k] = self.process_value(v)
            return value
        return self.process_scalar_value(value)

    def _set_dict(self, cfg, pointer, value):
        # pointer has at least one key
        if len(pointer) == 1:
            if pointer[0] in cfg:
                cfg[pointer[0]] = [cfg[pointer[0]], self.process_value(value)]
            else:
                cfg[pointer[0]] = self.process_value(value)
        else:
            if pointer[0] not in cfg:
                cfg[pointer[0]] = {}
            self._set_dict(cfg[pointer[0]], pointer[1:], self.process_value(value))



import os
from typing import Any, Optional

DEFAULT_URL = "http://localhost"


class NetboxStorageOpts:
    def __init__(self, url: str, token: str, insecure: bool = False, exact_host_filter: bool = False):
        self.url = url
        self.token = token
        self.insecure = insecure
        self.exact_host_filter = exact_host_filter

    @classmethod
    def parse_params(cls, conf_params: Optional[dict[str, str]], cli_opts: Any):
        url = os.getenv("NETBOX_URL") or conf_params.get("url") or DEFAULT_URL
        token = os.getenv("NETBOX_TOKEN", "").strip() or conf_params.get("token") or ""
        insecure = False
        if insecure_env := os.getenv("NETBOX_INSECURE", "").lower():
            insecure = insecure_env in ("true", "1", "t")
        else:
            insecure = bool(conf_params.get("insecure") or False)
        if exact_host_filter_env := os.getenv("NETBOX_EXACT_HOST_FILTER", "").lower():
            exact_host_filter = exact_host_filter_env in ("true", "1", "t")
        else:
            exact_host_filter = bool(conf_params.get("exact_host_filter") or False)
        return cls(url=url, token=token, insecure=insecure, exact_host_filter=exact_host_filter)



from typing import List

from annet.generators import BaseGenerator
from annet.storage import Storage
from . import policy_generator


def get_generators(store: Storage) -> List[BaseGenerator]:
    return policy_generator.get_generators(store)



from annet.annlib.rulebook import common
from annet.annlib.types import Op


def default_diff(old, new, diff_pre, _pops=(Op.AFFECTED,)):
    diff = common.default_diff(old, new, diff_pre, _pops)
    return diff


def ordered_diff(old, new, diff_pre, _pops=(Op.AFFECTED,)):
    diff = common.ordered_diff(old, new, diff_pre, _pops)
    return diff



from collections.abc import Iterator, Iterable
from dataclasses import dataclass
from enum import Enum
from typing import Generic, TypeVar, Sequence, Union, Any


class ConditionOperator(Enum):
    EQ = "=="
    GE = ">="
    GT = ">"
    LE = "<="
    LT = "<"
    BETWEEN_INCLUDED = "BETWEEN_INCLUDED"

    HAS = "has"
    HAS_ANY = "has_any"

    CUSTOM = "custom"


ValueT = TypeVar("ValueT")


@dataclass(frozen=True)
class SingleCondition(Generic[ValueT]):
    field: str
    operator: ConditionOperator
    value: ValueT

    def __and__(self, other: "Condition") -> "Condition":
        return AndCondition(self, other)

    def merge(self, other: "SingleCondition[Any]") -> "SingleCondition[Any]":
        if other.field != self.field:
            raise ValueError(f"Cannot merge conditions with different fields: {self.field} != {other.field}")
        if self.operator is ConditionOperator.LE:
            if other.operator is ConditionOperator.GE:
                return SingleCondition(self.field, ConditionOperator.BETWEEN_INCLUDED, (other.value, self.value))
            elif other.operator is ConditionOperator.LE:
                return SingleCondition(self.field, ConditionOperator.LE, min(self.value, other.value))
        elif self.operator is ConditionOperator.GE:
            if other.operator is ConditionOperator.LE:
                return other.merge(self)
            elif other.operator is ConditionOperator.GE:
                return SingleCondition(self.field, ConditionOperator.GE, max(self.value, other.value))
        elif self.operator is ConditionOperator.LT:
            if other.operator is ConditionOperator.LT:
                return SingleCondition(self.field, ConditionOperator.LT, min(self.value, other.value))
        elif self.operator is ConditionOperator.GT:
            if other.operator is ConditionOperator.GT:
                return SingleCondition(self.field, ConditionOperator.GT, max(self.value, other.value))
        raise ValueError(f"Cannot merge condition with operator {self.operator} and {other.operator}")


Condition = Union[SingleCondition, "AndCondition"]


class AndCondition:
    def __init__(self, *conditions: Condition):
        self.conditions: list[SingleCondition[Any]] = []
        for c in conditions:
            self.conditions.extend(self._unpack(c))

    def _unpack(self, other: Condition) -> Sequence[SingleCondition]:
        if isinstance(other, AndCondition):
            return other.conditions
        return [other]

    def __and__(self, other: Condition) -> "AndCondition":
        return AndCondition(*self.conditions, other)

    def __iadd__(self, other):
        self.conditions.extend(self._unpack(other))

    def __repr__(self):
        conditions = ", ".join(repr(c) for c in self.conditions)
        return f"AndCondition({conditions})"

    def __getitem__(self, item: str) -> SingleCondition[Any]:
        return next(c for c in self.conditions if c.field == item)

    def __contains__(self, item: str) -> bool:
        return any(c.field == item for c in self.conditions)

    def __len__(self) -> int:
        return len(self.conditions)

    def __iter__(self) -> Iterator[SingleCondition[Any]]:
        return iter(self.conditions)

    def find_all(self, item: str) -> Iterable[SingleCondition[Any]]:
        for condition in self.conditions:
            if condition.field == item:
                yield condition



from collections import OrderedDict as odict

from annet.annlib.rbparser import syntax


def config(config_tree, rules):
    implicit_config_tree = odict()
    for (row, rule) in rules.items():
        matched_lines = [line for line in config_tree.keys() if rule["regexp"].match(line)]
        if rule["type"] != "ignore":
            if not any(matched_lines) and row not in config_tree:
                implicit_config_tree[row] = odict()
        for line in matched_lines:
            implicit_config_tree[line] = config(config_tree[line], rule["children"])
    return implicit_config_tree


def compile_rules(device):
    return compile_tree(_implicit_tree(device))


def compile_tree(tree):
    rules = odict()
    for (_, attrs) in tree.items():
        rule = {
            "type": attrs["type"],
            "children": compile_tree(attrs["children"]) if attrs.get("children") else odict(),
            "regexp": syntax.compile_row_regexp(attrs["row"]),
        }
        rules[attrs["row"]] = rule
    return rules


def _implicit_tree(device):
    text = ""
    if device.hw.Huawei:
        if device.hw.Huawei.CE:
            text = """
                stp mode mstp
                stp enable
                undo ntp server disable
                undo telnet server disable
                undo dhcp enable
                !user-interface con *
                    user privilege level 3
                !user-interface vty ~
                    protocol inbound all
                netconf
                port link-flap trigger error-down
            """
        elif device.hw.Huawei.NE:
            text = """
                 !bgp *
                     !ipv4-family unicast
                         undo synchronization
                     !ipv6-family unicast
                         undo synchronization
                 !user-interface con *
                     user privilege level 3
                 !user-interface vty ~
                     protocol inbound all
                 aaa
                    undo user-password complexity-check
                 netconf
                 """
        else:
            text = """
                stp mode mstp
                !interface X?GigabitEthernet*
                    bpdu enable
                netconf
            """
    elif device.hw.Arista:
        # This part of configuration will not be visible in configuration
        text = r"""
                ip load-sharing trident fields ipv6 destination-port source-ip ingress-interface destination-ip source-port flow-label
                ip load-sharing trident fields ip source-ip source-port destination-ip destination-port ingress-interface
        """
    elif device.hw.Nexus:
        text = r"""
                # This part of configuration will not be visible in configuration if enabled
                snmp-server enable traps link linkDown
                snmp-server enable traps link linkUp
        """
        if device.hw.Nexus.N3x.N3432 \
            or (device.hw.Nexus.N9x.N9500 and "spine1" in device.tags) \
            or device.hw.Nexus.N9x.N9316 or device.hw.Cisco.Nexus.N9x.N9364:
            text += r"""
                # SVI
                !interface Vlan*
                    !shutdown
                !interface mgmt[0-9]*
                    no shutdown
                    mtu 1500
                !interface Ethernet1*
                    no shutdown
                # Лупбеки
                !interface */Loopback[0-9.]+/
                    no shutdown
                # Агрегаты
                !interface */port-channel[0-9.]+/
                    no shutdown
                # BGP
                !router bgp *
                    !neighbor *
                        no shutdown
            """

        elif device.hw.Nexus.N3x:
            # Cisco Nexus has some specific related to "shutdown" command
            # Behavior is cheked on Cisco Nexus 3132Q 6.0(2)U6(7)
            text += r"""
                # SVI
                !interface Vlan*
                    !shutdown
                !interface mgmt[0-9]*
                    no shutdown
                # Physical and NOT splitted interfaces and subifs
                !interface */Ethernet1\/[0-9.]*/
                    no shutdown
                # Physical and NOT splitted interfaces and subifs
                !interface */Ethernet1\/[0-9]+\/[0-9.]+/
                    # только explicit
                # Loopbacks
                !interface */Loopback[0-9.]+/
                    no shutdown
                # Port-Channels
                !interface */port-channel[0-9.]+/
                    no shutdown
                # BGP
                !router bgp *
                    !neighbor *
                        no shutdown
            """
    elif device.hw.Cisco:
        text += r"""
            !interface */\S*Ethernet\S+/
                mtu 1500
                no shutdown
            !interface */Loopback[0-9.]+/
                mtu 1500
                no shutdown
            !interface */port-channel[0-9.]+/
                mtu 1500
                no shutdown
        """
        if device.hw.Cisco.Catalyst:
            # this configuration is not visible in running-config when enabled
            text += r"""
                # line console aaa config
                !line con 0
                    authorization exec default
            """
    return parse_text(text)


def parse_text(text):
    return syntax.parse_text(text, {})



from dataclasses import dataclass
from enum import Enum
from typing import Generic, Sequence, Callable, Optional, TypeVar, Any

from .condition import SingleCondition, ConditionOperator, AndCondition


class MatchField(str, Enum):
    community = "community"
    extcommunity = "extcommunity"
    rd = "rd"
    interface = "interface"
    protocol = "protocol"
    net_len = "net_len"
    local_pref = "local_pref"
    metric = "metric"
    family = "family"

    as_path_length = "as_path_length"
    as_path_filter = "as_path_filter"
    ipv6_prefix = "ipv6_prefix"
    ip_prefix = "ip_prefix"


ValueT = TypeVar("ValueT")
_ConditionMethod = Callable[["ConditionFactory[ValueT]", ValueT], SingleCondition[ValueT]]


def condition_method(operator: ConditionOperator) -> _ConditionMethod:
    def method(self: "ConditionFactory[ValueT]", other: ValueT) -> SingleCondition[ValueT]:
        if operator.value not in self.supported_ops:
            raise NotImplementedError(f"Operator {operator.value} is not supported for field {self.field}")
        return SingleCondition(self.field, operator, other)

    method.__name__ = operator.value
    return method


class ConditionFactory(Generic[ValueT]):
    def __init__(self, field: str, supported_ops: list[str]):
        self.field = field
        self.supported_ops = supported_ops

    # https://github.com/python/typeshed/issues/3685
    eq = __eq__ = condition_method(ConditionOperator.EQ)  # type: ignore[assignment]
    gt = __gt__ = condition_method(ConditionOperator.GT)
    ge = __ge__ = condition_method(ConditionOperator.GE)
    lt = __lt__ = condition_method(ConditionOperator.LT)
    le = __le__ = condition_method(ConditionOperator.LE)
    between_included = condition_method(ConditionOperator.BETWEEN_INCLUDED)


class SetConditionFactory(Generic[ValueT]):
    def __init__(self, field: str) -> None:
        self.field = field

    def has(self, *values: ValueT) -> SingleCondition[Sequence[ValueT]]:
        return SingleCondition(self.field, ConditionOperator.HAS, values)

    def has_any(self, *values: ValueT) -> SingleCondition[Sequence[ValueT]]:
        return SingleCondition(self.field, ConditionOperator.HAS_ANY, values)


@dataclass(frozen=True)
class PrefixMatchValue:
    names: Sequence[str]
    or_longer: Optional[tuple[int, int]]  # ????


class Checkable:
    def __init__(self):
        self.community = SetConditionFactory[str](MatchField.community)
        self.extcommunity = SetConditionFactory[str](MatchField.extcommunity)
        self.rd = SetConditionFactory[str](MatchField.rd)
        self.interface = ConditionFactory[str](MatchField.interface, ["=="])
        self.protocol = ConditionFactory[str](MatchField.protocol, ["=="])
        self.net_len = ConditionFactory[int](MatchField.net_len, ["==", "!="])
        self.local_pref = ConditionFactory[int](MatchField.local_pref, ["<"])
        self.metric = ConditionFactory[int](MatchField.metric, ["=="])
        self.family = ConditionFactory[int](MatchField.family, ["=="])
        self.as_path_length = ConditionFactory[int](MatchField.as_path_length, ["==", ">=", "<=", "BETWEEN_INCLUDED"])

    def as_path_filter(self, name: str) -> SingleCondition[str]:
        return SingleCondition(MatchField.as_path_filter, ConditionOperator.EQ, name)

    def match_v6(self, *names: str, or_longer: Optional[tuple[int, int]] = None) -> SingleCondition[PrefixMatchValue]:
        return SingleCondition(MatchField.ipv6_prefix, ConditionOperator.CUSTOM, PrefixMatchValue(names, or_longer))

    def match_v4(self, *names: str, or_longer: Optional[tuple[int, int]] = None) -> SingleCondition[PrefixMatchValue]:
        return SingleCondition(MatchField.ip_prefix, ConditionOperator.CUSTOM, PrefixMatchValue(names, or_longer))


def merge_conditions(and_condition: AndCondition) -> AndCondition:
    conditions: dict[str, SingleCondition[Any]] = {}
    for condition in and_condition.conditions:
        if condition.field in conditions:
            conditions[condition.field] = conditions[condition.field].merge(condition)
        else:
            conditions[condition.field] = condition
    return AndCondition(*conditions.values())


R = Checkable()



from annet.adapters.netbox.common.models import NetboxDevice
from annet.rpl import R, RouteMap, Route

routemap = RouteMap[NetboxDevice]()


@routemap
def example1(device: NetboxDevice, route: Route):
    condition = (R.interface == "l0.0") & (R.protocol == "bgp")
    with route(condition, number=1, name="n1") as rule:
        rule.set_local_pref(100)
        rule.set_metric(100)
        rule.add_metric(200)
        rule.community.set("COMMUNITY_EXAMPLE_ADD")
        rule.as_path.set(12345, "123456")
        rule.allow()
    with route(R.protocol == "bgp", R.community.has("comm_name"), number=2, name="n2") as rule:
        rule.set_local_pref(100)
        rule.add_metric(200)
        rule.community.add("COMMUNITY_EXAMPLE_ADD")
        rule.community.remove("COMMUNITY_EXAMPLE_REMOVE")
        rule.allow()


@routemap
def example2(device: NetboxDevice, route: Route):
    with route(R.as_path_filter("ASP_EXAMPLE"), number=3, name="n3") as rule:
        rule.deny()
    with route(R.match_v6("IPV6_LIST_EXAMPLE"), number=4, name="n4") as rule:
        rule.allow()

    with route(R.as_path_length >= 1, R.as_path_length <= 20, number=4, name="n4") as rule:
        rule.allow()



from typing import Annotated, Optional, Union

from annet.bgp_models import Family, Redistribute
from .basemodel import BaseMeshModel, Concat, DictMerge, Merge, KeyDefaultDict
from .peer_models import MeshPeerGroup


class Aggregate(BaseMeshModel):
    policy: str
    routes: Annotated[tuple[str, ...], Concat()]
    export_policy: str
    as_path: str
    reference: str
    suppress: bool
    discard: bool
    as_set: bool


class FamilyOptions(BaseMeshModel):
    def __init__(self, **kwargs):
        kwargs.setdefault("aggregate", Aggregate())
        super().__init__(**kwargs)
    family: Family
    vrf_name: str
    multipath: int = 0
    global_multipath: int
    aggregate: Annotated[Aggregate, Merge()]
    redistributes: Annotated[tuple[Redistribute, ...], Concat()]
    allow_default: bool
    aspath_relax: bool
    igp_ignore: bool
    next_hop_policy: bool
    rib_import_policy: bool
    advertise_l2vpn_evpn: bool
    rib_group: bool
    loops: int
    advertise_bgp_static: bool


class _FamiliesMixin:
    def __init__(self, **kwargs):
        kwargs.setdefault("ipv4_unicast", FamilyOptions(family="ipv4_unicast"))
        kwargs.setdefault("ipv6_unicast", FamilyOptions(family="ipv6_unicast"))
        kwargs.setdefault("ipv4_labeled_unicast", FamilyOptions(family="ipv4_labeled_unicast"))
        kwargs.setdefault("ipv6_labeled_unicast", FamilyOptions(family="ipv6_labeled_unicast"))
        super().__init__(**kwargs)
    ipv4_unicast: Annotated[FamilyOptions, Merge()]
    ipv6_unicast: Annotated[FamilyOptions, Merge()]
    ipv4_labeled_unicast: Annotated[FamilyOptions, Merge()]
    ipv6_labeled_unicast: Annotated[FamilyOptions, Merge()]


class VrfOptions(_FamiliesMixin, BaseMeshModel):
    def __init__(self, vrf_name: str, **kwargs):
        kwargs.setdefault("ipv4_unicast", FamilyOptions(family="ipv4_unicast", vrf_name=vrf_name))
        kwargs.setdefault("ipv6_unicast", FamilyOptions(family="ipv6_unicast", vrf_name=vrf_name))
        kwargs.setdefault("ipv4_labeled_unicast", FamilyOptions(family="ipv4_labeled_unicast", vrf_name=vrf_name))
        kwargs.setdefault("ipv6_labeled_unicast", FamilyOptions(family="ipv6_labeled_unicast", vrf_name=vrf_name))
        kwargs.setdefault("groups", KeyDefaultDict(lambda x: MeshPeerGroup(name=x)))
        super().__init__(vrf_name=vrf_name, **kwargs)

    vrf_name: str
    vrf_name_global: Optional[str]
    import_policy: Optional[str]
    export_policy: Optional[str]
    rt_import: Annotated[tuple[str, ...], Concat()]
    rt_export: Annotated[tuple[str, ...], Concat()]
    rt_import_v4: Annotated[tuple[str, ...], Concat()]
    rt_export_v4: Annotated[tuple[str, ...], Concat()]
    route_distinguisher: Optional[str]
    static_label: Optional[int]  # FIXME: str?
    groups: Annotated[dict[str, MeshPeerGroup], DictMerge(Merge())]


class GlobalOptionsDTO(_FamiliesMixin, BaseMeshModel):
    def __init__(self, **kwargs):
        kwargs.setdefault("groups", KeyDefaultDict(lambda x: MeshPeerGroup(name=x)))
        kwargs.setdefault("vrf", KeyDefaultDict(lambda x: VrfOptions(vrf_name=x)))
        super().__init__(**kwargs)

    local_as: Union[int, str]
    loops: int
    multipath: int
    router_id: str
    vrf: Annotated[dict[str, VrfOptions], DictMerge(Merge())]
    groups: Annotated[dict[str, MeshPeerGroup], DictMerge(Merge())]



from annet.deploy import Fetcher
from annet.connectors import AdapterWithConfig
from typing import Dict, List, Any
from annet.storage import Device


class StubFetcher(Fetcher, AdapterWithConfig):
    @classmethod
    def with_config(cls, **kwargs: Dict[str, Any]) -> Fetcher:
        return cls(**kwargs)

    def fetch_packages(self, devices: List[Device],
                       processes: int = 1, max_slots: int = 0):
        raise NotImplementedError()

    def fetch(self, devices: List[Device],
              files_to_download: Dict[str, List[str]] = None,
              processes: int = 1, max_slots: int = 0):
        raise NotImplementedError()



import re

from annet.annlib.lib import cisco_collapse_vlandb as collapse_vlandb
from annet.annlib.lib import cisco_expand_vlandb as expand_vlandb
from annet.annlib.types import Op


VLANDB_CHUNK = 15
SWTRUNK_CHUNK = 5


# =====
def simple(rule, key, diff, hw, **_):
    yield from _process_vlandb(rule, key, diff, hw, False, VLANDB_CHUNK)


def swtrunk(rule, key, diff, hw, **_):
    yield from _process_vlandb(rule, key, diff, hw, True, SWTRUNK_CHUNK)


# =====
def _process_vlandb(rule, key, diff, hw, explicit_changing, multi_chunk):
    # pylint: disable=unused-argument
    for affected in diff[Op.AFFECTED]:
        # Изменилось содержимое блока vlan
        yield (True, affected["row"], affected["children"])

    (prefix, new, new_blocks) = _parse_vlancfg_actions(diff[Op.ADDED])
    (prefix2, old, old_blocks) = _parse_vlancfg_actions(diff[Op.REMOVED])
    if not prefix:
        prefix = prefix2

    if len(diff[Op.ADDED]) == 1 and len(new) == 0:
        # switchport trunk allowed vlan none
        yield (True, "%s none" % prefix, None)
        return
    for vlan_id in ((set(old_blocks.keys()) - set(new_blocks)) & new):
        # Удалено содержимое блока vlan, но сам влан остался
        yield (True, "%s %s" % (prefix, vlan_id), old_blocks[vlan_id])

    removed = old.difference(new)
    added = new.difference(old)
    if hw.Catalyst:
        # Каталисты не перечисляют вланы в batch режиме, если они представлены как блоки
        added -= new_blocks.keys()

    if removed:
        collapsed = collapse_vlandb(removed, hw.Catalyst)
        for chunk in _chunked(collapsed, multi_chunk):
            yield (False, "no %s%s%s" % (prefix, " remove " if explicit_changing else " ", ",".join(chunk)), None)

    if added:
        collapsed = collapse_vlandb(added, hw.Catalyst)
        for chunk in _chunked(collapsed, multi_chunk):
            yield (True, "%s%s%s" % (prefix, " add " if explicit_changing else " ", ",".join(chunk)), None)
    if new_blocks:
        for vlan_id, block in new_blocks.items():
            yield (True, "%s %s" % (prefix, vlan_id), block)


def _chunked(items, size):
    for offset in range(0, len(items), size):
        yield items[offset:offset + size]


def _parse_vlancfg_actions(actions):
    prefix = None
    vlandb = set()
    blocks = {}
    for action in actions:
        (prefix, part) = _parse_vlancfg(action["row"])
        if action["children"]:
            assert len(part) == 1, "vlandb block must contain one and only one vlanid: %s" % action["row"]
            blocks[list(part)[0]] = action["children"]
        vlandb.update(part)
    return (prefix, vlandb, blocks)


def _parse_vlancfg(row):
    # иногда циски ставят пробелы между влан ренджами, а иногда нет.
    words = re.sub(r",\s+", ",", row).split()

    if words[-1] == "none":
        # switchport trunk allowed vlan none
        return (" ".join(words[:-1]), set())
    assert re.match(r"[\d,-]+$", words[-1]), "Unable to parse vlancfg row: %s" % row
    prefix = " ".join(words[:-2] if words[-2] == "add" else words[:-1])
    vlancfg = words[-1]
    vlandb = expand_vlandb(vlancfg)
    return (prefix, vlandb)
